import os
import re
from copy import deepcopy
from lxml import etree
from datetime import datetime, timezone
from spatialist import Raster
from spatialist.ancillary import finder
from statistics import mean
from cesard.metadata.mapping import ASSET_MAP, NS_MAP
from cesard.metadata.extract import get_header_size, evaluate_types
from typing import Any
import logging

log = logging.getLogger('cesard')


def parse(
        meta: dict[str, dict[str, Any]],
        target: str,
        assets: list[str],
        exist_ok: bool = False
) -> None:
    """
    Wrapper for :func:`~cesard.metadata.xml.source_xml` and :func:`~cesard.metadata.xml.product_xml`.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    target:
        A path pointing to the root directory of a product scene.
    assets:
        List of paths to all GeoTIFF and VRT assets of the currently processed ARD product.
    exist_ok:
        Do not create files if they already exist?
    """
    evaluate_types(meta)
    
    platform = meta['common']['platformShortName']
    pid = {'Sentinel-1': 's1', 'ERS': 'ers', 'ENVISAT': 'env'}[platform]
    key = f"{pid}-{meta['prod']['productName-short'].lower()}"
    nsmap = deepcopy(NS_MAP)
    nsmap[key] = nsmap.pop('placeholder').format(meta['common']['constellation'])
    src_url = nsmap[key].replace('spec', key.split('-')[1]).replace('role', 'source')
    prod_url = nsmap[key].replace('spec', key.split('-')[1]).replace('role', 'product')
    
    nsmap.update({key: src_url})
    source_xml(meta=meta, target=target, nsmap=nsmap, ard_ns=key, exist_ok=exist_ok)
    nsmap.update({key: prod_url})
    product_xml(meta=meta, target=target, assets=assets, nsmap=nsmap, ard_ns=key,
                exist_ok=exist_ok)


def source_xml(
        meta: dict[str, dict[str, Any]],
        target: str,
        nsmap: dict[str, str],
        ard_ns: str,
        exist_ok: bool = False
) -> None:
    """
    Function to generate source-level metadata for an ARD product in `OGC 10-157r4` compliant XML format.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    target:
        A path pointing to the root directory of a product scene.
    nsmap:
        Dictionary listing abbreviation (key) and URI (value) of all necessary XML namespaces.
    ard_ns:
        Abbreviation of the ARD namespace. E.g., `s1-nrb` for the NRB ARD product.
    exist_ok:
        Do not create files if they already exist?
    """
    metadir = os.path.join(target, 'source')
    for uid in list(meta['source'].keys()):
        scene = os.path.basename(meta['source'][uid]['filename']).split('.')[0]
        outname = os.path.join(metadir, '{}.xml'.format(scene))
        if os.path.isfile(outname) and exist_ok:
            continue
        log.info(f'creating {os.path.relpath(outname, target)}')
        timeStart = meta['source'][uid]['timeStart'].isoformat()
        timeStop = meta['source'][uid]['timeStop'].isoformat()
        
        root = etree.Element(_nsc('_:EarthObservation', nsmap, ard_ns=ard_ns), nsmap=nsmap,
                             attrib={_nsc('gml:id', nsmap): scene + '_1'})
        _om_time(root=root, nsmap=nsmap, scene_id=scene, time_start=timeStart, time_stop=timeStop)
        _om_procedure(root=root, nsmap=nsmap, ard_ns=ard_ns, scene_id=scene, meta=meta, uid=uid, prod=False)
        observedProperty = etree.SubElement(root, _nsc('om:observedProperty', nsmap),
                                            attrib={'nilReason': 'inapplicable'})
        _om_feature_of_interest(root=root, nsmap=nsmap, scene_id=scene,
                                extent=meta['source'][uid]['geom_xml_envelope'],
                                center=meta['source'][uid]['geom_xml_center'])
        
        ################################################################################################################
        result = etree.SubElement(root, _nsc('om:result', nsmap))
        earthObservationResult = etree.SubElement(result, _nsc('eop:EarthObservationResult', nsmap),
                                                  attrib={_nsc('gml:id', nsmap): scene + '_9'})
        product = etree.SubElement(earthObservationResult, _nsc('eop:product', nsmap))
        productInformation = etree.SubElement(product, _nsc('_:ProductInformation', nsmap, ard_ns=ard_ns))
        fileName = etree.SubElement(productInformation, _nsc('eop:fileName', nsmap))
        serviceReference = etree.SubElement(fileName, _nsc('ows:ServiceReference', nsmap),
                                            attrib={_nsc('xlink:href', nsmap): scene})
        requestMessage = etree.SubElement(serviceReference, _nsc('ows:RequestMessage', nsmap))
        
        org_src_files_dir = os.path.join(metadir, uid)
        if os.path.isdir(org_src_files_dir):
            org_src_files = finder(target=org_src_files_dir, matchlist=['*.safe', '*.xml'], foldermode=0)
            if len(org_src_files) > 0:
                for file in org_src_files:
                    href = './' + os.path.relpath(file, metadir).replace('\\', '/')
                    product = etree.SubElement(earthObservationResult, _nsc('eop:product', nsmap))
                    productInformation = etree.SubElement(product, _nsc('_:ProductInformation', nsmap, ard_ns=ard_ns))
                    fileName = etree.SubElement(productInformation, _nsc('eop:fileName', nsmap))
                    serviceReference = etree.SubElement(fileName, _nsc('ows:ServiceReference', nsmap),
                                                        attrib={_nsc('xlink:href', nsmap): href})
                    requestMessage = etree.SubElement(serviceReference, _nsc('ows:RequestMessage', nsmap))
                    dataFormat = etree.SubElement(productInformation, _nsc('_:dataFormat', nsmap, ard_ns=ard_ns))
                    dataFormat.text = 'XML'
        ################################################################################################################
        metaDataProperty = etree.SubElement(root, _nsc('eop:metaDataProperty', nsmap))
        earthObservationMetaData = etree.SubElement(metaDataProperty, _nsc('_:EarthObservationMetaData', nsmap,
                                                                           ard_ns=ard_ns))
        
        identifier = etree.SubElement(earthObservationMetaData, _nsc('eop:identifier', nsmap))
        identifier.text = scene
        doi = etree.SubElement(earthObservationMetaData, _nsc('eop:doi', nsmap))
        doi.text = meta['source'][uid]['doi']
        acquisitionType = etree.SubElement(earthObservationMetaData, _nsc('eop:acquisitionType', nsmap))
        acquisitionType.text = meta['source'][uid]['acquisitionType']
        status = etree.SubElement(earthObservationMetaData, _nsc('eop:status', nsmap))
        status.text = meta['source'][uid]['status']
        
        processing = etree.SubElement(earthObservationMetaData, _nsc('eop:processing', nsmap))
        processingInformation = etree.SubElement(processing, _nsc('_:ProcessingInformation', nsmap, ard_ns=ard_ns))
        
        fields = [(processingInformation, 'eop:processingCenter', {'codeSpace': 'urn:esa:eop:Sentinel1:facility'},
                   'processingCenter'),
                  (processingInformation, 'eop:processingDate', None, 'processingDate'),
                  (processingInformation, 'eop:processorName', None, 'processorName'),
                  (processingInformation, 'eop:processorVersion', None, 'processorVersion'),
                  (processingInformation, 'eop:processingMode', None, 'processingMode'),
                  (processingInformation, '_:orbitDataSource', None, 'orbitDataSource'),
                  (processingInformation, '_:orbitStateVector', {'access': meta['source'][uid]['orbitDataAccess']},
                   'orbitStateVector'),
                  (processingInformation, '_:lutApplied', None, 'lutApplied'),
                  (earthObservationMetaData, '_:productType', {'codeSpace': 'urn:esa:eop:Sentinel1:class'},
                   'productType'),
                  (earthObservationMetaData, '_:dataGeometry', None, 'dataGeometry'),
                  (earthObservationMetaData, '_:azimuthPixelSpacing', {'uom': 'm'}, 'azimuthPixelSpacing'),
                  (earthObservationMetaData, '_:rangePixelSpacing', {'uom': 'm'}, 'rangePixelSpacing'),
                  (earthObservationMetaData, '_:meanFaradayRotationAngle', {'uom': 'deg'}, 'faradayMeanRotationAngle'),
                  (earthObservationMetaData, '_:referenceFaradayRotation',
                   {_nsc('xlink:href', nsmap): str(meta['source'][uid]['faradayRotationReference'])}, None),
                  (earthObservationMetaData, '_:ionosphereIndicator', None, 'ionosphereIndicator')]
        
        for parent, field_dst, attrib, field_src in fields:
            element = etree.SubElement(_parent=parent,
                                       _tag=_nsc(text=field_dst, nsmap=nsmap, ard_ns=ard_ns),
                                       attrib=attrib)
            if field_src is not None:
                value = meta['source'][uid][field_src]
                if field_src == 'processingDate':
                    value = value.isoformat()
                elif field_src == 'orbitDataSource':
                    value = value.upper()
                elif field_src in ['azimuthPixelSpacing', 'rangePixelSpacing']:
                    value = str(mean(value.values()))
                elif field_src == 'processorVersion':
                    value = value[meta['source'][uid]['processorName']]
                element.text = value
        
        processingLevel = etree.SubElement(processingInformation, _nsc('_:processingLevel', nsmap, ard_ns=ard_ns))
        processingLevel.text = meta['common']['processingLevel']
        
        for swath in meta['source'][uid]['swaths']:
            fields = [
                (processingInformation, 'azimuthLookBandwidth', {'uom': 'Hz', 'beam': swath}),
                (processingInformation, 'rangeLookBandwidth', {'uom': 'Hz', 'beam': swath}),
                (earthObservationMetaData, 'azimuthNumberOfLooks', {'beam': swath}),
                (earthObservationMetaData, 'rangeNumberOfLooks', {'beam': swath}),
                (earthObservationMetaData, 'azimuthResolution', {'uom': 'm', 'beam': swath}),
                (earthObservationMetaData, 'rangeResolution', {'uom': 'm', 'beam': swath})
            ]
            
            for parent, key, attrib in fields:
                if key in meta['source'][uid].keys():
                    element = etree.SubElement(_parent=parent,
                                               _tag=_nsc(f'_:{key}', nsmap, ard_ns=ard_ns),
                                               attrib=attrib)
                    element.text = str(meta['source'][uid][key][swath])
        
        performance = etree.SubElement(earthObservationMetaData, _nsc('_:performance', nsmap, ard_ns=ard_ns))
        performanceIndicators = etree.SubElement(performance, _nsc('_:PerformanceIndicators', nsmap, ard_ns=ard_ns))
        noiseEquivalentIntensityType = etree.SubElement(performanceIndicators,
                                                        _nsc('_:noiseEquivalentIntensityType', nsmap, ard_ns=ard_ns),
                                                        attrib={'uom': 'dB'})
        noiseEquivalentIntensityType.text = str(meta['source'][uid]['perfNoiseEquivalentIntensityType'])
        
        for pol in meta['common']['polarisationChannels']:
            for type in ['minimum', 'mean', 'maximum']:
                estimate = etree.SubElement(_parent=performanceIndicators,
                                            _tag=_nsc('_:estimates', nsmap, ard_ns=ard_ns),
                                            attrib={'pol': pol, 'type': type})
                estimate.text = str(meta['source'][uid]['perfEstimates'][pol][type])
        
        fields = [(performanceIndicators, '_:equivalentNumberOfLooks', 'perfEquivalentNumberOfLooks'),
                  (performanceIndicators, '_:peakSideLobeRatio', 'perfPeakSideLobeRatio'),
                  (performanceIndicators, '_:integratedSideLobeRatio', 'perfIntegratedSideLobeRatio')]
        
        for parent, field_dst, field_src in fields:
            element = etree.SubElement(parent,
                                       _nsc(field_dst, nsmap, ard_ns=ard_ns))
            element.text = str(meta['source'][uid][field_src])
        
        polCalMatrices = etree.SubElement(earthObservationMetaData, _nsc('_:polCalMatrices', nsmap, ard_ns=ard_ns),
                                          attrib={
                                              _nsc('xlink:href', nsmap): str(meta['source'][uid]['polCalMatrices'])})
        ################################################################################################################
        etree.indent(root)
        tree = etree.ElementTree(root)
        tree.write(outname, pretty_print=True, xml_declaration=True, encoding='utf-8')


def product_xml(
        meta: dict[str, dict[str, Any]],
        target: str,
        assets: list[str],
        nsmap: dict[str, str],
        ard_ns: str,
        exist_ok: bool = False
) -> None:
    """
    Function to generate product-level metadata for an ARD product in `OGC 10-157r4` compliant XML format.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    target:
        A path pointing to the root directory of a product scene.
    assets:
        List of paths to all GeoTIFF and VRT assets of the currently processed ARD product.
    nsmap:
        Dictionary listing abbreviation (key) and URI (value) of all necessary XML namespaces.
    ard_ns:
        Abbreviation of the ARD namespace. E.g., `s1-nrb` for the NRB ARD product.
    exist_ok:
        Do not create files if they already exist?
    """
    scene_id = os.path.basename(target)
    outname = os.path.join(target, '{}.xml'.format(scene_id))
    if os.path.isfile(outname) and exist_ok:
        return
    log.info(f'creating {os.path.relpath(outname, target)}')
    timeCreated = meta['prod']['timeCreated'].isoformat()
    timeStart = meta['prod']['timeStart'].isoformat()
    timeStop = meta['prod']['timeStop'].isoformat()
    
    root = etree.Element(_nsc('_:EarthObservation', nsmap, ard_ns=ard_ns), nsmap=nsmap,
                         attrib={_nsc('gml:id', nsmap): scene_id + '_1'})
    _om_time(root=root, nsmap=nsmap, scene_id=scene_id, time_start=timeStart, time_stop=timeStop)
    _om_procedure(root=root, nsmap=nsmap, ard_ns=ard_ns, scene_id=scene_id, meta=meta, prod=True)
    observedProperty = etree.SubElement(root, _nsc('om:observedProperty', nsmap),
                                        attrib={'nilReason': 'inapplicable'})
    _om_feature_of_interest(root=root, nsmap=nsmap, scene_id=scene_id,
                            extent=meta['prod']['geom_xml_envelope'],
                            center=meta['prod']['geom_xml_center'])
    
    ####################################################################################################################
    result = etree.SubElement(root, _nsc('om:result', nsmap))
    earthObservationResult = etree.SubElement(result, _nsc('eop:EarthObservationResult', nsmap),
                                              attrib={_nsc('gml:id', nsmap): scene_id + '_9'})
    product = etree.SubElement(earthObservationResult, _nsc('eop:product', nsmap))
    productInformation = etree.SubElement(product, _nsc('_:ProductInformation', nsmap, ard_ns=ard_ns))
    fileName = etree.SubElement(productInformation, _nsc('eop:fileName', nsmap))
    serviceReference = etree.SubElement(fileName, _nsc('ows:ServiceReference', nsmap),
                                        attrib={_nsc('xlink:href', nsmap): scene_id})
    requestMessage = etree.SubElement(serviceReference, _nsc('ows:RequestMessage', nsmap))
    
    for asset in assets:
        relpath = './' + os.path.relpath(asset, target).replace('\\', '/')
        
        no_data = None
        header_size = None
        data_format = 'VRT'
        byte_order = None
        data_type = None
        z_error = None
        if asset.endswith('.tif'):
            with Raster(asset) as ras:
                no_data = str(ras.nodata)
            header_size = str(get_header_size(tif=asset))
            data_format = 'COG'
            byte_order = 'little-endian'
            data_type = 'FLOAT 32'
            prefix = '[0-9a-z]{5}-'
            match = re.search(prefix + f"({'|'.join(meta['prod']['compression_zerrors'].keys())})",
                              os.path.basename(asset))
            if match is not None:
                k = match.group()
                k = k.removeprefix(re.search(prefix, k).group())
                z_error = str(meta['prod']['compression_zerrors'][k])
        
        product = etree.SubElement(earthObservationResult, _nsc('eop:product', nsmap))
        productInformation = etree.SubElement(product, _nsc('_:ProductInformation', nsmap, ard_ns=ard_ns))
        fileName = etree.SubElement(productInformation, _nsc('eop:fileName', nsmap))
        serviceReference = etree.SubElement(fileName, _nsc('ows:ServiceReference', nsmap),
                                            attrib={_nsc('xlink:href', nsmap): relpath})
        requestMessage = etree.SubElement(serviceReference, _nsc('ows:RequestMessage', nsmap))
        
        size = etree.SubElement(productInformation, _nsc('eop:size', nsmap), attrib={'uom': 'bytes'})
        size.text = str(os.path.getsize(asset))
        
        lookup = [
            ('headerSize', header_size, {'uom': 'bytes'}),
            ('byteOrder', byte_order, None),
            ('dataFormat', data_format, None),
            ('noDataValue', no_data, None)
        ]
        for key, value, attrib in lookup:
            if value is not None:
                element = etree.SubElement(_parent=productInformation,
                                           _tag=_nsc(text=f'_:{key}', nsmap=nsmap, ard_ns=ard_ns),
                                           attrib=attrib)
                element.text = value
        
        if data_type is not None:
            dataType = etree.SubElement(productInformation, _nsc('_:dataType', nsmap, ard_ns=ard_ns))
            dataType.text = data_type.split()[0]
            bitsPerSample = etree.SubElement(productInformation, _nsc('_:bitsPerSample', nsmap, ard_ns=ard_ns))
            bitsPerSample.text = data_type.split()[1]
        if z_error is not None:
            compressionType = etree.SubElement(productInformation, _nsc('_:compressionType', nsmap, ard_ns=ard_ns))
            compressionType.text = meta['prod']['compression_type']
            compressionzError = etree.SubElement(productInformation, _nsc('_:compressionZError', nsmap, ard_ns=ard_ns))
            compressionzError.text = z_error
        
        if 'annotation' in asset:
            pattern = '(dm|ei|em|lc|ld|li|gs|id|np-[vh]{2}|sg|wm).tif'
            key = re.search(pattern, asset).groups()[0][:2]
            
            sampleType = etree.SubElement(productInformation, _nsc('_:sampleType', nsmap, ard_ns=ard_ns),
                                          attrib={'uom': 'unitless' if ASSET_MAP[key]['unit'] is None else
                                          ASSET_MAP[key]['unit']})
            sampleType.text = ASSET_MAP[key]['type']
            
            if key in ['dm', 'id']:
                dataType.text = 'UINT'
                bitsPerSample.text = '8'
                
                if key == 'dm':
                    with Raster(asset) as dm_ras:
                        band_descr = [dm_ras.raster.GetRasterBand(band).GetDescription() for band in
                                      range(1, dm_ras.bands + 1)]
                    samples = [x for x in band_descr if x in ASSET_MAP[key]['allowed']]
                    for i, sample in enumerate(samples):
                        bitValue = etree.SubElement(productInformation, _nsc('_:bitValue', nsmap, ard_ns=ard_ns),
                                                    attrib={'band': str(i + 1),
                                                            'name': sample})
                        bitValue.text = '1'
                else:  # key == 'id'
                    src_list = list(meta['source'].keys())
                    src_target = [os.path.basename(meta['source'][src]['filename']).replace('.SAFE',
                                                                                            '').replace('.zip', '')
                                  for src in src_list]
                    for i, s in enumerate(src_target):
                        bitValue = etree.SubElement(productInformation, _nsc('_:bitValue', nsmap, ard_ns=ard_ns),
                                                    attrib={'band': '1', 'name': s})
                        bitValue.text = str(i + 1)
            
            if key == 'ei':
                ellipsoidalHeight = etree.SubElement(productInformation, _nsc('_:ellipsoidalHeight', nsmap,
                                                                              ard_ns=ard_ns),
                                                     attrib={'uom': 'm'})
                ellipsoidalHeight.text = meta['prod']['ellipsoidalHeight']
        
        if 'measurement' in asset and not asset.endswith('.vrt'):
            creationTime = etree.SubElement(productInformation, _nsc('_:creationTime', nsmap, ard_ns=ard_ns))
            creationTime.text = datetime.fromtimestamp(os.path.getctime(asset), tz=timezone.utc).isoformat()
            polarisation = etree.SubElement(productInformation, _nsc('_:polarisation', nsmap, ard_ns=ard_ns))
            polarisation.text = re.search('-[vh]{2}', relpath).group().removeprefix('-').upper()
            numBorderPixels = etree.SubElement(productInformation, _nsc('_:numBorderPixels', nsmap, ard_ns=ard_ns))
            numBorderPixels.text = str(meta['prod']['numBorderPixels'])
    
    ####################################################################################################################
    metaDataProperty = etree.SubElement(root, _nsc('eop:metaDataProperty', nsmap))
    earthObservationMetaData = etree.SubElement(metaDataProperty, _nsc('_:EarthObservationMetaData', nsmap,
                                                                       ard_ns=ard_ns))
    
    identifier = etree.SubElement(earthObservationMetaData, _nsc('eop:identifier', nsmap))
    identifier.text = scene_id
    if meta['prod']['doi'] is not None:
        doi = etree.SubElement(earthObservationMetaData, _nsc('eop:doi', nsmap))
        doi.text = meta['prod']['doi']
    acquisitionType = etree.SubElement(earthObservationMetaData, _nsc('eop:acquisitionType', nsmap))
    acquisitionType.text = meta['prod']['acquisitionType']
    status = etree.SubElement(earthObservationMetaData, _nsc('eop:status', nsmap))
    status.text = meta['prod']['status']
    
    processing = etree.SubElement(earthObservationMetaData, _nsc('eop:processing', nsmap))
    processingInformation = etree.SubElement(processing, _nsc('_:ProcessingInformation', nsmap, ard_ns=ard_ns))
    if meta['prod']['processingCenter'] is not None:
        processingCenter = etree.SubElement(processingInformation, _nsc('eop:processingCenter', nsmap),
                                            attrib={'codeSpace': 'urn:esa:eop:Sentinel1:facility'})
        processingCenter.text = meta['prod']['processingCenter']
    processingDate = etree.SubElement(processingInformation, _nsc('eop:processingDate', nsmap))
    processingDate.text = timeCreated
    processorName = etree.SubElement(processingInformation, _nsc('eop:processorName', nsmap))
    processorName.text = meta['prod']['processorName']
    processorVersion = etree.SubElement(processingInformation, _nsc('eop:processorVersion', nsmap))
    processorVersion.text = meta['prod']['processorVersion'][meta['prod']['processorName']]
    processingMode = etree.SubElement(processingInformation, _nsc('eop:processingMode', nsmap),
                                      attrib={'codeSpace': 'urn:esa:eop:Sentinel1:class'})
    processingMode.text = meta['prod']['processingMode']
    processingLevel = etree.SubElement(processingInformation, _nsc('_:processingLevel', nsmap, ard_ns=ard_ns))
    processingLevel.text = meta['common']['processingLevel']
    for src in list(meta['source'].keys()):
        src_path = '{}.xml'.format(os.path.basename(meta['source'][src]['filename']).split('.')[0])
        src_target = os.path.join('./source', src_path).replace('\\', '/')
        sourceProduct = etree.SubElement(processingInformation, _nsc('_:sourceProduct', nsmap, ard_ns=ard_ns),
                                         attrib={_nsc('xlink:href', nsmap): src_target})
    auxData1 = etree.SubElement(processingInformation, _nsc('_:auxiliaryDataSetFileName', nsmap, ard_ns=ard_ns),
                                attrib={_nsc('xlink:href', nsmap): meta['prod']['ancillaryData_KML']})
    speckleFilterApplied = etree.SubElement(processingInformation, _nsc('_:speckleFilterApplied', nsmap,
                                                                        ard_ns=ard_ns))
    speckleFilterApplied.text = str(meta['prod']['speckleFilterApplied']).lower()
    noiseRemovalApplied = etree.SubElement(processingInformation, _nsc('_:noiseRemovalApplied', nsmap, ard_ns=ard_ns))
    noiseRemovalApplied.text = str(meta['prod']['noiseRemovalApplied']).lower()
    value = meta['prod']['noiseRemovalAlgorithm']
    if meta['prod']['noiseRemovalApplied']:
        noiseRemovalAlgorithm = etree.SubElement(processingInformation,
                                                 _nsc('_:noiseRemovalAlgorithm', nsmap, ard_ns=ard_ns),
                                                 attrib={_nsc('xlink:href', nsmap): value})
    if meta['prod']['RTCAlgorithm'] is not None:
        rtcAlgorithm = etree.SubElement(processingInformation, _nsc('_:RTCAlgorithm', nsmap, ard_ns=ard_ns),
                                        attrib={_nsc('xlink:href', nsmap): meta['prod']['RTCAlgorithm']})
    key = 'windNormBackscatterMeasurement'
    if hasattr(meta['prod'], key) and meta['prod'][key] is not None:
        windNormBackscatterMeasurement = etree.SubElement(processingInformation,
                                                          _nsc('_:windNormBackscatterMeasurement',
                                                               nsmap, ard_ns=ard_ns))
        windNormBackscatterMeasurement.text = meta['prod']['windNormBackscatterMeasurement']
        windNormBackscatterConvention = etree.SubElement(processingInformation,
                                                         _nsc('_:windNormBackscatterConvention', nsmap, ard_ns=ard_ns))
        windNormBackscatterConvention.text = meta['prod']['windNormBackscatterConvention']
        windNormReferenceDirection = etree.SubElement(processingInformation,
                                                      _nsc('_:windNormReferenceDirection', nsmap, ard_ns=ard_ns),
                                                      attrib={'uom': 'deg'})
        windNormReferenceDirection.text = str(meta['prod']['windNormReferenceDirection'])
        
        windNormReferenceModel = etree.SubElement(processingInformation, _nsc('_:windNormReferenceModel', nsmap,
                                                                              ard_ns=ard_ns),
                                                  attrib={_nsc('xlink:href', nsmap):
                                                              meta['prod']['windNormReferenceModel']})
        windNormReferenceSpeed = etree.SubElement(processingInformation,
                                                  _nsc('_:windNormReferenceSpeed', nsmap, ard_ns=ard_ns),
                                                  attrib={'uom': 'm_s'})
        windNormReferenceSpeed.text = str(meta['prod']['windNormReferenceSpeed'])
        windNormReferenceType = etree.SubElement(processingInformation,
                                                 _nsc('_:windNormReferenceType', nsmap, ard_ns=ard_ns))
        windNormReferenceType.text = meta['prod']['windNormReferenceType']
    
    geoCorrAlgorithm = etree.SubElement(processingInformation, _nsc('_:geoCorrAlgorithm', nsmap, ard_ns=ard_ns),
                                        attrib={_nsc('xlink:href', nsmap): str(meta['prod']['geoCorrAlgorithm'])})
    geoCorrResamplingMethod = etree.SubElement(processingInformation, _nsc('_:geoCorrResamplingAlgorithm', nsmap,
                                                                           ard_ns=ard_ns))
    geoCorrResamplingMethod.text = meta['prod']['geoCorrResamplingMethod'].upper()
    demReference = etree.SubElement(processingInformation, _nsc('_:DEMReference', nsmap, ard_ns=ard_ns),
                                    attrib={'name': meta['prod']['demName'],
                                            'dem': meta['prod']['demType'],
                                            _nsc('xlink:href', nsmap): meta['prod']['demReference']})
    demResamplingMethod = etree.SubElement(processingInformation, _nsc('_:DEMResamplingMethod', nsmap, ard_ns=ard_ns))
    demResamplingMethod.text = meta['prod']['demResamplingMethod'].upper()
    demAccess = etree.SubElement(processingInformation, _nsc('_:DEMAccess', nsmap, ard_ns=ard_ns),
                                 attrib={_nsc('xlink:href', nsmap): meta['prod']['demAccess']})
    demGSD = etree.SubElement(processingInformation, _nsc('_:DEMGroundSamplingDistance', nsmap, ard_ns=ard_ns),
                              attrib={'uom': meta['prod']['demGSD'].split()[1]})
    demGSD.text = meta['prod']['demGSD'].split()[0]
    egmReference = etree.SubElement(processingInformation, _nsc('_:EGMReference', nsmap, ard_ns=ard_ns),
                                    attrib={_nsc('xlink:href', nsmap): meta['prod']['demEGMReference']})
    egmResamplingMethod = etree.SubElement(processingInformation, _nsc('_:EGMResamplingMethod', nsmap,
                                                                       ard_ns=ard_ns))
    egmResamplingMethod.text = meta['prod']['demEGMResamplingMethod'].upper()
    
    productType = etree.SubElement(earthObservationMetaData, _nsc('_:productType', nsmap, ard_ns=ard_ns),
                                   attrib={'codeSpace': 'urn:esa:eop:Sentinel1:class'})
    productType.text = meta['prod']['productName-short']
    refDoc = etree.SubElement(earthObservationMetaData, _nsc('_:refDoc', nsmap, ard_ns=ard_ns),
                              attrib={'name': meta['prod']['productName'],
                                      'version': meta['prod']['card4l-version'],
                                      _nsc('xlink:href', nsmap): meta['prod']['card4l-link']})
    lookup = [
        ('azimuthNumberOfLooks', None),
        ('rangeNumberOfLooks', None),
        ('equivalentNumberOfLooks', None),
        ('radiometricAccuracyRelative', {'uom': 'dB'}),
        ('radiometricAccuracyAbsolute', {'uom': 'dB'})
    ]
    for key, attrib in lookup:
        element = etree.SubElement(_parent=earthObservationMetaData,
                                   _tag=_nsc(text=f'_:{key}', nsmap=nsmap, ard_ns=ard_ns),
                                   attrib=attrib)
        element.text = str(meta['prod'][key])
    
    radacc_ref = str(meta['prod']['radiometricAccuracyReference'])
    radiometricAccuracyReference = etree.SubElement(earthObservationMetaData,
                                                    _nsc('_:radiometricAccuracyReference', nsmap, ard_ns=ard_ns),
                                                    attrib={_nsc('xlink:href', nsmap): radacc_ref})
    
    lookup = [('geoCorrAccuracyType', None)]
    lookup += [(f'geoCorrAccuracy{direction}{measure}', {'uom': 'm'})
               for direction in ['Northern', 'Eastern']
               for measure in ['STDev', 'Bias']]
    lookup += [('geoCorrAccuracy_rRMSE', {'uom': 'm'})]
    for key, attrib in lookup:
        element = etree.SubElement(_parent=earthObservationMetaData,
                                   _tag=_nsc(f'_:{key}', nsmap, ard_ns=ard_ns),
                                   attrib=attrib)
        element.text = str(meta['prod'][key])
    
    geoacc_ref = meta['prod']['geoCorrAccuracyReference']
    if geoacc_ref is not None:
        geoCorrAccuracyReference = etree.SubElement(earthObservationMetaData,
                                                    _tag=_nsc('_:geoCorrAccuracyReference', nsmap, ard_ns=ard_ns),
                                                    attrib={_nsc('xlink:href', nsmap): geoacc_ref})
    
    lookup = [
        ('numLines', None, None),
        ('numPixelsPerLine', None, None),
        ('pxSpacingColumn', 'columnSpacing', {'uom': 'm'}),
        ('pxSpacingRow', 'rowSpacing', {'uom': 'm'}),
        ('pixelCoordinateConvention', None, None),
        ('backscatterMeasurement', None, None),
        ('backscatterConvention', None, None),
        ('backscatterConversionEq', None, {'uom': 'dB'})
    ]
    for key_src, key_dst, attrib in lookup:
        if key_dst is None:
            key_dst = key_src
        element = etree.SubElement(_parent=earthObservationMetaData,
                                   _tag=_nsc(text=f'_:{key_dst}', nsmap=nsmap, ard_ns=ard_ns),
                                   attrib=attrib)
        element.text = str(meta['prod'][key_src])
    
    griddingConvention = etree.SubElement(_parent=earthObservationMetaData,
                                          _tag=_nsc('_:griddingConvention', nsmap, ard_ns=ard_ns),
                                          attrib={_nsc('xlink:href', nsmap):
                                                      meta['prod']['griddingConventionURL']})
    
    lookup = [
        ('mgrsID', None),
        ('crsEPSG', {'codeSpace': 'urn:esa:eop:crs'}),
        ('crsWKT', None)
    ]
    for key, attrib in lookup:
        element = etree.SubElement(_parent=earthObservationMetaData,
                                   _tag=_nsc(text=f'_:{key}', nsmap=nsmap, ard_ns=ard_ns),
                                   attrib=attrib)
        element.text = str(meta['prod'][key])
    ####################################################################################################################
    etree.indent(root)
    tree = etree.ElementTree(root)
    tree.write(outname, pretty_print=True, xml_declaration=True, encoding='utf-8')


def _nsc(
        text: str,
        nsmap: dict[str, str],
        ard_ns: str | None = None
) -> str:
    ns, key = text.split(':')
    if ard_ns is not None and ns == '_':
        ns = ard_ns
    return '{{{0}}}{1}'.format(nsmap[ns], key)


def _om_time(
        root: etree.Element,
        nsmap: dict[str, str],
        scene_id: str,
        time_start: str,
        time_stop: str
) -> None:
    """
    Creates the `om:phenomenonTime` and `om:resultTime` XML elements.
    
    Parameters
    ----------
    root:
        Root XML element.
    nsmap:
        Dictionary listing abbreviation (key) and URI (value) of all necessary XML namespaces.
    scene_id:
        Scene basename.
    time_start:
        Start time of the scene acquisition.
    time_stop:
        Stop time of the acquisition.
    """
    phenomenonTime = etree.SubElement(root, _nsc('om:phenomenonTime', nsmap))
    timePeriod = etree.SubElement(phenomenonTime, _nsc('gml:TimePeriod', nsmap),
                                  attrib={_nsc('gml:id', nsmap): scene_id + '_2'})
    beginPosition = etree.SubElement(timePeriod, _nsc('gml:beginPosition', nsmap))
    beginPosition.text = time_start
    endPosition = etree.SubElement(timePeriod, _nsc('gml:endPosition', nsmap))
    endPosition.text = time_stop
    
    resultTime = etree.SubElement(root, _nsc('om:resultTime', nsmap))
    timeInstant = etree.SubElement(resultTime, _nsc('gml:TimeInstant', nsmap),
                                   attrib={_nsc('gml:id', nsmap): scene_id + '_3'})
    timePosition = etree.SubElement(timeInstant, _nsc('gml:timePosition', nsmap))
    timePosition.text = time_stop


def _om_procedure(
        root: etree.Element,
        nsmap: dict[str, str],
        ard_ns: str,
        scene_id: str,
        meta: dict[str, dict[str, Any]],
        uid: str | None = None,
        prod: bool = True
) -> None:
    """
    Creates the `om:procedure/eop:EarthObservationEquipment` XML elements and all relevant subelements for source and
    product metadata. Differences between source and product are controlled using the `prod=[True|False]` switch.
    
    Parameters
    ----------
    root:
        Root XML element.
    nsmap:
        Dictionary listing abbreviation (key) and URI (value) of all necessary XML namespaces.
    ard_ns:
        Abbreviation of the ARD namespace. E.g., `s1-nrb` for the NRB ARD product.
    scene_id: str
        Scene basename.
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    uid:
        Unique identifier of a source SLC scene.
    prod:
        Return XML subelements for further usage in :func:`~cesard.metadata.xml.product_xml` parsing function?
        Default is True. If False, the XML subelements for further usage in the :func:`~cesard.metadata.xml.source_xml`
        parsing function will be returned.
    """
    procedure = etree.SubElement(root, _nsc('om:procedure', nsmap))
    earthObservationEquipment = etree.SubElement(procedure, _nsc('eop:EarthObservationEquipment', nsmap),
                                                 attrib={_nsc('gml:id', nsmap): scene_id + '_4'})
    
    # eop:platform
    platform0 = etree.SubElement(earthObservationEquipment, _nsc('eop:platform', nsmap))
    if prod:
        platform1 = etree.SubElement(platform0, _nsc('eop:Platform', nsmap))
    else:
        platform1 = etree.SubElement(platform0, _nsc('_:Platform', nsmap, ard_ns=ard_ns))
    shortName = etree.SubElement(platform1, _nsc('eop:shortName', nsmap))
    shortName.text = meta['common']['platformShortName'].upper()
    serialIdentifier = etree.SubElement(platform1, _nsc('eop:serialIdentifier', nsmap))
    serialIdentifier.text = meta['common']['platformIdentifier']
    if not prod:
        satReference = etree.SubElement(platform1, _nsc('_:satelliteReference', nsmap, ard_ns=ard_ns),
                                        attrib={_nsc('xlink:href', nsmap): meta['common']['platformReference']})
    
    # eop:instrument
    instrument0 = etree.SubElement(earthObservationEquipment, _nsc('eop:instrument', nsmap))
    instrument1 = etree.SubElement(instrument0, _nsc('eop:Instrument', nsmap))
    shortName = etree.SubElement(instrument1, _nsc('eop:shortName', nsmap))
    shortName.text = meta['common']['instrumentShortName']
    
    # eop:sensor
    sensor0 = etree.SubElement(earthObservationEquipment, _nsc('eop:sensor', nsmap))
    sensor1 = etree.SubElement(sensor0, _nsc('_:Sensor', nsmap, ard_ns=ard_ns))
    sensorType = etree.SubElement(sensor1, _nsc('eop:sensorType', nsmap))
    sensorType.text = meta['common']['sensorType']
    operationalMode = etree.SubElement(sensor1, _nsc('eop:operationalMode', nsmap),
                                       attrib={'codeSpace': 'urn:esa:eop:C-SAR:operationalMode'})
    operationalMode.text = meta['common']['operationalMode']
    swathIdentifier = etree.SubElement(sensor1, _nsc('eop:swathIdentifier', nsmap),
                                       attrib={'codeSpace': 'urn:esa:eop:C-SAR:swathIdentifier'})
    swathIdentifier.text = meta['common']['swathIdentifier']
    radarBand = etree.SubElement(sensor1, _nsc('_:radarBand', nsmap, ard_ns=ard_ns))
    radarBand.text = meta['common']['radarBand']
    if not prod:
        radarCenterFreq = etree.SubElement(sensor1, _nsc('_:radarCenterFrequency', nsmap, ard_ns=ard_ns),
                                           attrib={'uom': 'Hz'})
        radarCenterFreq.text = '{:.3e}'.format(meta['common']['radarCenterFreq'])
        value = meta['source'][uid]['sensorCalibration']
        if value is not None:
            sensorCalibration = etree.SubElement(sensor1, _nsc('_:sensorCalibration',
                                                               nsmap, ard_ns=ard_ns),
                                                 attrib={_nsc('xlink:href', nsmap): value})
    
    # eop:acquisitionParameters
    acquisitionParameters = etree.SubElement(earthObservationEquipment, _nsc('eop:acquisitionParameters', nsmap))
    acquisition = etree.SubElement(acquisitionParameters, _nsc('_:Acquisition', nsmap, ard_ns=ard_ns))
    orbitNumber = etree.SubElement(acquisition, _nsc('eop:orbitNumber', nsmap))
    orbitNumber.text = str(meta['common']['orbitNumber_abs'])
    orbitDirection = etree.SubElement(acquisition, _nsc('eop:orbitDirection', nsmap))
    orbitDirection.text = meta['common']['orbitDirection'].upper()
    wrsLongitudeGrid = etree.SubElement(acquisition, _nsc('eop:wrsLongitudeGrid', nsmap),
                                        attrib={'codeSpace': 'urn:esa:eop:Sentinel1:relativeOrbits'})
    wrsLongitudeGrid.text = meta['common']['wrsLongitudeGrid']
    if not prod:
        value = meta['source'][uid]['ascendingNodeDate']
        ascendingNodeDate = etree.SubElement(acquisition, _nsc('eop:ascendingNodeDate', nsmap))
        if value is not None:
            ascendingNodeDate.text = value.isoformat()
        startTimeFromAscendingNode = etree.SubElement(acquisition, _nsc('eop:startTimeFromAscendingNode', nsmap),
                                                      attrib={'uom': 'ms'})
        startTimeFromAscendingNode.text = meta['source'][uid]['timeStartFromAscendingNode']
        completionTimeFromAscendingNode = etree.SubElement(acquisition,
                                                           _nsc('eop:completionTimeFromAscendingNode', nsmap),
                                                           attrib={'uom': 'ms'})
        completionTimeFromAscendingNode.text = meta['source'][uid]['timeCompletionFromAscendingNode']
        instrumentAzimuthAngle = etree.SubElement(acquisition, _nsc('eop:instrumentAzimuthAngle', nsmap),
                                                  attrib={'uom': 'deg'})
        instrumentAzimuthAngle.text = str(meta['source'][uid]['instrumentAzimuthAngle'])
    polarisationMode = etree.SubElement(acquisition, _nsc('sar:polarisationMode', nsmap))
    polarisationMode.text = meta['common']['polarisationMode']
    polarisationChannels = etree.SubElement(acquisition, _nsc('sar:polarisationChannels', nsmap))
    polarisationChannels.text = ', '.join(meta['common']['polarisationChannels'])
    if prod:
        numberOfAcquisitions = etree.SubElement(acquisition, _nsc('_:numberOfAcquisitions', nsmap, ard_ns=ard_ns))
        numberOfAcquisitions.text = meta['prod']['numberOfAcquisitions']
    else:
        antennaLookDirection = etree.SubElement(acquisition, _nsc('sar:antennaLookDirection', nsmap))
        antennaLookDirection.text = meta['common']['antennaLookDirection']
        minimumIncidenceAngle = etree.SubElement(acquisition, _nsc('sar:minimumIncidenceAngle', nsmap),
                                                 attrib={'uom': 'deg'})
        minimumIncidenceAngle.text = str(meta['source'][uid]['incidenceAngleMin'])
        maximumIncidenceAngle = etree.SubElement(acquisition, _nsc('sar:maximumIncidenceAngle', nsmap),
                                                 attrib={'uom': 'deg'})
        maximumIncidenceAngle.text = str(meta['source'][uid]['incidenceAngleMax'])
        orbitMeanAltitude = etree.SubElement(acquisition, _nsc('_:orbitMeanAltitude', nsmap, ard_ns=ard_ns),
                                             attrib={'uom': 'm'})
        orbitMeanAltitude.text = meta['common']['orbitMeanAltitude']
        if meta['common']['platformShortName'] == 'Sentinel':
            dataTakeID = etree.SubElement(acquisition, _nsc('_:dataTakeID', nsmap, ard_ns=ard_ns))
            dataTakeID.text = meta['source'][uid]['datatakeID']
        majorCycleID = etree.SubElement(acquisition, _nsc('_:majorCycleID', nsmap, ard_ns=ard_ns))
        majorCycleID.text = meta['source'][uid]['majorCycleID']


def _om_feature_of_interest(
        root: etree.Element,
        nsmap: dict[str, str],
        scene_id: str,
        extent: str,
        center: str
):
    """
    Creates the `om:featureOfInterest` XML elements.
    
    Parameters
    ----------
    root:
        Root XML element.
    nsmap:
        Dictionary listing abbreviation (key) and URI (value) of all necessary XML namespaces.
    scene_id:
        Scene basename.
    extent:
        Footprint coordinates of the scene.
    center:
        Center coordinates of the footprint.
    """
    featureOfInterest = etree.SubElement(root, _nsc('om:featureOfInterest', nsmap))
    footprint = etree.SubElement(featureOfInterest, _nsc('eop:Footprint', nsmap),
                                 attrib={_nsc('gml:id', nsmap): scene_id + '_5'})
    
    multiExtentOf = etree.SubElement(footprint, _nsc('eop:multiExtentOf', nsmap))
    multiSurface = etree.SubElement(multiExtentOf, _nsc('gml:MultiSurface', nsmap),
                                    attrib={_nsc('gml:id', nsmap): scene_id + '_6'})
    surfaceMember = etree.SubElement(multiSurface, _nsc('gml:surfaceMember', nsmap))
    polygon = etree.SubElement(surfaceMember, _nsc('gml:Polygon', nsmap),
                               attrib={_nsc('gml:id', nsmap): scene_id + '_7'})
    exterior = etree.SubElement(polygon, _nsc('gml:exterior', nsmap))
    linearRing = etree.SubElement(exterior, _nsc('gml:LinearRing', nsmap))
    posList = etree.SubElement(linearRing, _nsc('gml:posList', nsmap))
    posList.text = extent
    
    centerOf = etree.SubElement(footprint, _nsc('eop:centerOf', nsmap))
    point = etree.SubElement(centerOf, _nsc('gml:Point', nsmap), attrib={_nsc('gml:id', nsmap): scene_id + '_8'})
    pos = etree.SubElement(point, _nsc('gml:pos', nsmap))
    pos.text = center



# 'z_error': Maximum error threshold on values for LERC* compression.
# Will be ignored if a compression algorithm is used that isn't related to LERC.
LERC_ERR_THRES = {
    'vv-g-lin': 1e-4,
    'vh-g-lin': 1e-4,
    'hh-g-lin': 1e-4,
    'hv-g-lin': 1e-4,
    'vv-s-lin': 1e-4,
    'vh-s-lin': 1e-4,
    'hh-s-lin': 1e-4,
    'hv-s-lin': 1e-4,
    'ei': 1e-3,
    'em': 1e-3,
    'dm': 0.0,
    'li': 1e-2,
    'lc': 0.1,
    'ld': 1e-3,
    'gs': 1e-4,
    'id': 0.0,
    'np-vv': 2e-5,
    'np-vh': 2e-5,
    'np-hh': 2e-5,
    'np-hv': 2e-5,
    'sg': 1e-4,
    'wm': 2e-5
}

# XML namespaces are identifiers, and it is not their goal to be directly usable for schema retrieval:
# https://stackoverflow.com/a/30761004
NS_MAP = {
    'placeholder': 'http://earth.esa.int/{0}/spec/role/1.0',
    'sar': 'http://www.opengis.net/sar/2.1',
    'eop': 'http://www.opengis.net/eop/2.1',
    'om': 'http://www.opengis.net/om/2.0',
    'gml': 'http://www.opengis.net/gml/3.2',
    'ows': 'http://www.opengis.net/ows/2.0',
    'xlink': 'http://www.w3.org/1999/xlink'
}

ASSET_MAP = {
    'dm': {'type': 'Mask',
           'unit': None,
           'role': 'data-mask',
           'title': 'Data Mask Image',
           'allowed': ['not layover, nor shadow',
                       'layover',
                       'shadow',
                       'layover and shadow',
                       'ocean',
                       'lakes',
                       'rivers']},
    'ei': {'type': 'Angle',
           'unit': 'deg',
           'role': 'ellipsoid-incidence-angle',
           'title': 'Ellipsoid Incidence Angle'},
    'em': {'type': 'Elevation',
           'unit': 'meters',
           'role': 'digital-elevation-model',
           'title': 'Digital Elevation Model'},
    'lc': {'type': 'Scattering Area',
           'unit': 'square_meters',
           'role': 'contributing-area',
           'title': 'Local Contributing Area'},
    'ld': {'type': 'Angle',
           'unit': 'deg',
           'role': 'range-look-direction-angle',
           'title': 'Range Look Direction Angle'},
    'li': {'type': 'Angle',
           'unit': 'deg',
           'role': 'local-incidence-angle',
           'title': 'Local Incidence Angle'},
    'gs': {'type': 'Ratio',
           'unit': None,
           'role': 'gamma-sigma-ratio',
           'title': 'Gamma0 RTC to sigma0 RTC ratio'},
    'id': {'type': 'AcqID',
           'unit': None,
           'role': 'acquisition-id',
           'title': 'Acquisition ID Image'},
    'np': {'type': 'Sigma-0',
           'unit': 'dB',
           'role': 'noise-power',
           'title': 'Noise Power'},
    'sg': {'type': 'Ratio',
           'unit': None,
           'role': 'sigma-gamma-ratio',
           'title': 'Sigma0 RTC to gamma0 RTC ratio'},
    'wm': {'type': 'Sigma-0',
           'unit': None,
           'role': 'wind-modelled-backscatter',
           'title': 'wind-modelled backscatter (OCN CMOD NRCS)'}
}

URL = {
    'ancillaryData_KML': 'https://sentiwiki.copernicus.eu/__attachments/1692737/'
                         'S2A_OPER_GIP_TILPAR_MPC__20151209T095117_V20150622T000000_21000101T000000_B00.zip',
    'card4l_nrb': 'https://ceos.org/ard/files/PFS/NRB/v5.5/CARD4L-PFS_NRB_v5.5.pdf',
    'card4l_orb': 'https://ceos.org/ard/files/PFS/ORB/v1.0/'
                  'CARD4L_Product_Family_Specification_Ocean_Radar_Backscatter-v1.0.pdf',
    'egm2008': 'https://doi.org/10.1029/2011JB008916',
    'griddingConventionURL': 'https://www.mgrs-data.org/data/documents/nga_mgrs_doc.pdf',
    'platformReference': {
        'envisat-1': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=2',
        'ers-1': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=220',
        'ers-2': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=221',
        'sentinel-1a': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=575',
        'sentinel-1b': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=576',
        'sentinel-1c': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=577',
        'sentinel-1d': 'https://database.eohandbook.com/database/missionsummary.aspx?missionID=814'
    }
}

DEM_MAP = {
    'GETASSE30':
        {'access': 'https://step.esa.int/auxdata/dem/GETASSE30',
         'ref': 'https://www.brockmann-consult.de/beam/doc/help/visat/GETASSE30ElevationModel.html',
         'type': 'elevation',
         'gsd': '30 arcsec',
         'egm': None},
    'Copernicus 10m EEA DEM':
        {'access': 'ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_EEA-10-DGED/2021_1',
         'ref': 'https://doi.org/10.5270/ESA-c5d3d65',
         'type': 'surface',
         'gsd': '10 m',
         'egm': URL['egm2008']},
    'Copernicus 30m Global DEM':
        {'access': 'https://copernicus-dem-30m.s3.eu-central-1.amazonaws.com/',
         'ref': 'https://copernicus-dem-30m.s3.amazonaws.com/readme.html',
         'type': 'surface',
         'gsd': '30 m',
         'egm': URL['egm2008']},
    'Copernicus 30m Global DEM II':
        {'access': 'ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_GLO-30-DGED/2021_1',
         'ref': 'https://doi.org/10.5270/ESA-c5d3d65',
         'type': 'surface',
         'gsd': '30 m',
         'egm': URL['egm2008']},
    'Copernicus 90m Global DEM II':
        {'access': 'ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_GLO-90-DGED/2021_1',
         'ref': 'https://doi.org/10.5270/ESA-c5d3d65',
         'type': 'surface',
         'gsd': '90 m',
         'egm': URL['egm2008']}
}



import os
import re
from statistics import mean
from datetime import datetime, timezone
import pystac
from pystac.extensions.sar import SarExtension, FrequencyBand, Polarization, ObservationDirection
from pystac.extensions.sat import SatExtension, OrbitState
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.view import ViewExtension
from pystac.extensions.mgrs import MgrsExtension
from pystac.extensions.file import FileExtension, ByteOrder
from pystac.extensions.raster import RasterExtension, RasterBand, DataType
from pystac.extensions.classification import ClassificationExtension, Classification
from spatialist import Raster
from spatialist.ancillary import finder
from cesard.metadata.mapping import ASSET_MAP
from cesard.metadata.extract import get_header_size, evaluate_types
from cesard.ancillary import compute_hash
from typing import Any
import logging

log = logging.getLogger('cesard')


def parse(
        meta: dict[str, dict[str, Any]],
        target: str,
        assets: list[str],
        exist_ok: bool = False
) -> None:
    """
    Wrapper for :func:`~cesard.metadata.stac.source_json` and :func:`~cesard.metadata.stac.product_json`.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    target:
        A path pointing to the root directory of a product scene.
    assets:
        List of paths to all GeoTIFF and VRT assets of the currently processed ARD product.
    exist_ok:
        Do not create files if they already exist?
    """
    evaluate_types(meta)
    source_json(meta=meta, target=target, exist_ok=exist_ok)
    product_json(meta=meta, target=target, assets=assets, exist_ok=exist_ok)


def source_json(
        meta: dict[str, dict[str, Any]],
        target: str,
        exist_ok: bool = False
) -> None:
    """
    Function to generate source-level metadata for an ARD product in STAC compliant JSON format.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    target:
        A path pointing to the root directory of a product scene.
    exist_ok:
        Do not create files if they already exist?
    """
    metadir = os.path.join(target, 'source')
    for uid in list(meta['source'].keys()):
        scene = os.path.splitext(os.path.basename(meta['source'][uid]['filename']))[0]
        outname = os.path.join(metadir, '{}.json'.format(scene))
        if os.path.isfile(outname) and exist_ok:
            continue
        log.info(f'creating {os.path.relpath(outname, target)}')
        start = meta['source'][uid]['timeStart']
        stop = meta['source'][uid]['timeStop']
        date = start + (stop - start) / 2
        #################################################################################
        # Initialise STAC item
        item = pystac.Item(id=scene,
                           geometry=meta['source'][uid]['geom_stac_geometry_4326'],
                           bbox=meta['source'][uid]['geom_stac_bbox_4326'],
                           datetime=date,
                           properties={})
        #################################################################################
        # Add common metadata
        item.common_metadata.start_datetime = start
        item.common_metadata.end_datetime = stop
        item.common_metadata.created = meta['source'][uid]['processingDate']
        item.common_metadata.instruments = [meta['common']['instrumentShortName'].lower()]
        item.common_metadata.constellation = meta['common']['constellation']
        item.common_metadata.platform = meta['common']['platformFullname']
        #################################################################################
        # Add properties
        
        # prepare values
        values = {}
        for key in ['rangeResolution', 'rangePixelSpacing', 'rangeNumberOfLooks',
                    'azimuthResolution', 'azimuthPixelSpacing', 'azimuthNumberOfLooks']:
            value = meta['source'][uid][key]
            values[key] = mean(value.values()) if isinstance(value, dict) else value
        for key in ['rangeLookBandwidth', 'azimuthLookBandwidth']:
            if hasattr(meta['source'][uid], key):
                value = meta['source'][uid][key]
                if isinstance(value, dict):
                    value = {k: v / 1e9 for k, v in value.items()}  # GHz
                if isinstance(value, (float, int)):
                    value = value / 1e9
                values[key] = value
        ####################################
        # sat extension properties
        sat_ext = SatExtension.ext(item, add_if_missing=True)
        sat_ext.apply(orbit_state=OrbitState[meta['common']['orbitDirection'].upper()],
                      relative_orbit=meta['common']['orbitNumber_rel'],
                      absolute_orbit=meta['common']['orbitNumber_abs'],
                      anx_datetime=meta['source'][uid]['ascendingNodeDate'])
        ####################################
        # sar extension properties
        sar_ext = SarExtension.ext(item, add_if_missing=True)
        sar_ext.apply(instrument_mode=meta['common']['operationalMode'],
                      frequency_band=FrequencyBand[meta['common']['radarBand'].upper()],
                      polarizations=[Polarization[pol] for pol in meta['common']['polarisationChannels']],
                      product_type=meta['source'][uid]['productType'],
                      center_frequency=float(meta['common']['radarCenterFreq'] / 1e9),
                      resolution_range=values['rangeResolution'],
                      resolution_azimuth=values['azimuthResolution'],
                      pixel_spacing_range=values['rangePixelSpacing'],
                      pixel_spacing_azimuth=values['azimuthPixelSpacing'],
                      looks_range=values['rangeNumberOfLooks'],
                      looks_azimuth=values['azimuthNumberOfLooks'],
                      looks_equivalent_number=meta['source'][uid]['perfEquivalentNumberOfLooks'],
                      observation_direction=ObservationDirection[meta['common']['antennaLookDirection']])
        ####################################
        # view extension properties
        view_ext = ViewExtension.ext(item, add_if_missing=True)
        view_ext.apply(incidence_angle=meta['source'][uid]['incidenceAngleMidSwath'],
                       azimuth=meta['source'][uid]['instrumentAzimuthAngle'])
        ####################################
        # processing extension properties
        item.stac_extensions.append('https://stac-extensions.github.io/processing/v1.1.0/schema.json')
        item.properties['processing:facility'] = meta['source'][uid]['processingCenter']
        item.properties['processing:software'] = meta['source'][uid]['processorVersion']
        item.properties['processing:level'] = meta['common']['processingLevel']
        ####################################
        # card4l extension properties
        item.stac_extensions.append('https://stac-extensions.github.io/card4l/v0.1.0/sar/source.json')
        item.properties['card4l:specification'] = meta['prod']['productName-short']
        item.properties['card4l:specification_version'] = meta['prod']['card4l-version']
        item.properties['card4l:beam_id'] = meta['common']['swathIdentifier']
        item.properties['card4l:orbit_mean_altitude'] = float(meta['common']['orbitMeanAltitude'])
        
        proc_param = {'lut_applied': meta['source'][uid]['lutApplied']}
        
        pairs = [('range_look_bandwidth', 'rangeLookBandwidth'),
                 ('azimuth_look_bandwidth', 'azimuthLookBandwidth')]
        for key_dst, key_src in pairs:
            if hasattr(values, key_src):
                proc_param[key_dst] = values[key_src]
        
        item.properties['card4l:source_processing_parameters'] = proc_param
        
        keys = [
            ('card4l:incidence_angle_far_range', 'incidenceAngleMax'),
            ('card4l:incidence_angle_near_range', 'incidenceAngleMin'),
            ('card4l:integrated_sidelobe_ratio', 'perfIntegratedSideLobeRatio'),
            ('card4l:ionosphere_indicator', 'ionosphereIndicator'),
            ('card4l:mean_faraday_rotation_angle', 'faradayMeanRotationAngle'),
            ('card4l:noise_equivalent_intensity', 'perfEstimates'),
            ('card4l:noise_equivalent_intensity_type', 'perfNoiseEquivalentIntensityType'),
            ('card4l:orbit_data_source', 'orbitDataSource'),
            ('card4l:peak_sidelobe_ratio', 'perfPeakSideLobeRatio'),
            ('card4l:resolution_range', 'rangeResolution'),
            ('card4l:resolution_azimuth', 'azimuthResolution'),
            ('card4l:source_geometry', 'dataGeometry')
        ]
        for key_dst, key_src in keys:
            val = meta['source'][uid][key_src]
            if val is not None:
                item.properties[key_dst] = val
        #################################################################################
        # Add links
        links = [
            {'rel': 'card4l-document',
             'target': meta['prod']['card4l-link'].replace('.pdf', '.docx'),
             'media_type': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
             'title': f'CARD4L Product Family Specification: '
                      f'{meta["prod"]["productName"]} (v{meta["prod"]["card4l-version"]})'},
            {'rel': 'card4l-document',
             'target': meta['prod']['card4l-link'],
             'media_type': 'application/pdf',
             'title': f'CARD4L Product Family Specification: '
                      f'{meta["prod"]["productName"]} (v{meta["prod"]["card4l-version"]})'},
            {'rel': 'about',
             'target': meta['source'][uid]['doi'],
             'media_type': None,
             'title': 'Product definition reference.'},
            {'rel': 'access',
             'target': meta['source'][uid]['access'],
             'media_type': None,
             'title': 'Product data access.'},
            {'rel': 'satellite',
             'target': meta['common']['platformReference'],
             'media_type': None,
             'title': 'CEOS Missions, Instruments and Measurements Database record'},
            {'rel': 'state-vectors',
             'target': meta['source'][uid]['orbitStateVector'],
             'media_type': None,
             'title': 'Orbit data file containing state vectors.'},
            {'rel': 'sensor-calibration',
             'target': meta['source'][uid]['sensorCalibration'],
             'media_type': None,
             'title': 'Reference describing sensor calibration parameters.'},
            {'rel': 'pol-cal-matrices',
             'target': meta['source'][uid]['polCalMatrices'],
             'media_type': None,
             'title': 'Reference to the complex-valued polarimetric distortion matrices.'},
            {'rel': 'referenced-faraday-rotation',
             'target': meta['source'][uid]['faradayRotationReference'],
             'media_type': None,
             'title': 'Reference describing the method used to derive the estimate '
                      'for the mean Faraday rotation angle.'}]
        
        for link in links:
            if link['target'] is not None:
                item.add_link(link=pystac.Link(**link))
        #################################################################################
        # Add assets
        relpath = os.path.relpath(outname.replace('.json', '.xml'), metadir)
        xml_relpath = './' + relpath.replace('\\', '/')
        item.add_asset(key='card4l',
                       asset=pystac.Asset(href=xml_relpath,
                                          title='Metadata in XML format.',
                                          media_type=pystac.MediaType.XML,
                                          roles=['metadata', 'card4l']))
        _asset_add_orig_src(metadir=metadir, uid=uid, item=item)
        #################################################################################
        # (validate) and save the result
        
        # item.validate()  # for later
        item.save_object(dest_href=outname)


def _asset_add_orig_src(
        metadir: str,
        uid: str,
        item: pystac.Item
) -> None:
    """
    Helper function to add the original source metadata files as assets to a STAC item.
    
    Parameters
    ----------
    metadir:
        Source directory of the current ARD product.
    uid:
        Unique identifier of a source scene.
    item:
        The pystac.Item to add the assets to.
    
    """
    pattern = r'^(.+?)-\d{8}t\d{6}-\d{8}t\d{6}-\w+-\w+-\d{3}\.xml$'
    prefixes = {'calibration': 'Calibration metadata',
                'noise': 'Estimated thermal noise look-up tables',
                'rfi': 'Radio Frequency Interference metadata'}
    
    root_dir = os.path.join(metadir, uid)
    if not os.path.isdir(root_dir):
        return
    file_list = finder(target=root_dir, matchlist=['*.safe', '*.xml'], foldermode=0)
    if len(file_list) > 0:
        for file in file_list:
            basename = os.path.basename(file)
            href = './' + os.path.relpath(file, metadir).replace('\\', '/')
            if basename == 'manifest.safe':
                key = 'manifest'
                title = 'Mandatory product metadata'
            else:
                try:
                    key = re.match(pattern, basename).group(1)
                except AttributeError:
                    raise RuntimeError(
                        'Unexpected file in original source metadata directory: ' + os.path.join(root_dir, file))
                title = prefixes.get(key.split('-')[0], 'Measurement metadata')
                if title == 'Measurement metadata':
                    title = title + ' ({},{})'.format(key.split('-')[1].upper(), key.split('-')[3].upper())
                else:
                    title = title + ' ({},{})'.format(key.split('-')[2].upper(), key.split('-')[4].upper())
            
            item.add_asset(key=key,
                           asset=pystac.Asset(href=href,
                                              title=title,
                                              media_type=pystac.MediaType.XML,
                                              roles=['metadata']))


def product_json(
        meta: dict[str, dict[str, Any]],
        target: str,
        assets: list[str],
        exist_ok: bool = False
):
    """
    Function to generate product-level metadata for an ARD product in STAC compliant JSON format.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    target:
        A path pointing to the root directory of a product scene.
    assets:
        List of paths to all GeoTIFF and VRT assets of the currently processed ARD product.
    exist_ok:
        Do not create files if they already exist?
    """
    scene_id = os.path.basename(target)
    outname = os.path.join(target, '{}.json'.format(scene_id))
    if os.path.isfile(outname) and exist_ok:
        return
    log.info(f'creating {os.path.relpath(outname, target)}')
    start = meta['prod']['timeStart']
    stop = meta['prod']['timeStop']
    date = start + (stop - start) / 2
    mgrs = meta['prod']['mgrsID']
    
    # Initialise STAC item
    item = pystac.Item(id=scene_id,
                       geometry=meta['prod']['geom_stac_geometry_4326'],
                       bbox=meta['prod']['geom_stac_bbox_4326'],
                       datetime=date,
                       properties={})
    
    # Add common metadata
    item.common_metadata.license = meta['prod']['licence']
    item.common_metadata.start_datetime = start
    item.common_metadata.end_datetime = stop
    item.common_metadata.created = meta['prod']['timeCreated']
    item.common_metadata.instruments = [meta['common']['instrumentShortName'].lower()]
    item.common_metadata.constellation = meta['common']['constellation']
    item.common_metadata.platform = meta['common']['platformFullname']
    item.common_metadata.gsd = float(meta['prod']['pxSpacingColumn'])
    
    # Initialise STAC extensions for properties
    sar_ext = SarExtension.ext(item, add_if_missing=True)
    sat_ext = SatExtension.ext(item, add_if_missing=True)
    proj_ext = ProjectionExtension.ext(item, add_if_missing=True)
    mgrs_ext = MgrsExtension.ext(item, add_if_missing=True)
    item.stac_extensions.append('https://stac-extensions.github.io/processing/v1.1.0/schema.json')
    item.stac_extensions.append('https://stac-extensions.github.io/card4l/v0.1.0/sar/product.json')
    #################################################################################
    # Add properties
    sat_ext.apply(orbit_state=OrbitState[meta['common']['orbitDirection'].upper()],
                  relative_orbit=meta['common']['orbitNumber_rel'],
                  absolute_orbit=meta['common']['orbitNumber_abs'])
    sar_ext.apply(instrument_mode=meta['common']['operationalMode'],
                  frequency_band=FrequencyBand[meta['common']['radarBand'].upper()],
                  polarizations=[Polarization[pol] for pol in meta['common']['polarisationChannels']],
                  product_type=meta['prod']['productName-short'],
                  looks_range=meta['prod']['rangeNumberOfLooks'],
                  looks_azimuth=meta['prod']['azimuthNumberOfLooks'],
                  looks_equivalent_number=meta['prod']['equivalentNumberOfLooks'])
    proj_ext.apply(epsg=int(meta['prod']['crsEPSG']),
                   wkt2=meta['prod']['crsWKT'],
                   bbox=meta['prod']['geom_stac_bbox_native'],
                   shape=[int(meta['prod']['numPixelsPerLine']), int(meta['prod']['numLines'])],
                   transform=meta['prod']['transform'])
    mgrs_ext.apply(latitude_band=mgrs[2:3],
                   grid_square=mgrs[3:],
                   utm_zone=int(mgrs[:2]))
    if meta['prod']['processingCenter'] is not None:
        item.properties['processing:facility'] = meta['prod']['processingCenter']
    item.properties['processing:software'] = meta['prod']['processorVersion']
    item.properties['processing:level'] = meta['common']['processingLevel']
    item.properties['card4l:specification'] = meta['prod']['productName-short']
    item.properties['card4l:specification_version'] = meta['prod']['card4l-version']
    item.properties['card4l:beam_id'] = meta['common']['swathIdentifier']
    item.properties['card4l:measurement_type'] = meta['prod']['backscatterMeasurement']
    item.properties['card4l:measurement_convention'] = meta['prod']['backscatterConvention']
    item.properties['card4l:pixel_coordinate_convention'] = meta['prod']['pixelCoordinateConvention']
    item.properties['card4l:speckle_filtering'] = meta['prod']['speckleFilterApplied']
    item.properties['card4l:noise_removal_applied'] = meta['prod']['noiseRemovalApplied']
    item.properties['card4l:conversion_eq'] = meta['prod']['backscatterConversionEq']
    item.properties['card4l:relative_radiometric_accuracy'] = float(meta['prod']['radiometricAccuracyRelative'])
    item.properties['card4l:absolute_radiometric_accuracy'] = float(meta['prod']['radiometricAccuracyAbsolute'])
    item.properties['card4l:resampling_method'] = meta['prod']['geoCorrResamplingMethod']
    item.properties['card4l:dem_resampling_method'] = meta['prod']['demResamplingMethod']
    item.properties['card4l:egm_resampling_method'] = meta['prod']['demEGMResamplingMethod']
    item.properties['card4l:gridding_convention'] = 'Sentinel-2 MGRS'
    item.properties['card4l:geometric_accuracy_type'] = meta['prod']['geoCorrAccuracyType']
    for x in ['Northern', 'Eastern']:
        key = ['geoCorrAccuracy{}{}'.format(x, y) for y in ['STDev', 'Bias']]
        stddev = float(meta['prod'][key[0]])
        bias = float(meta['prod'][key[1]])
        item.properties['card4l:{}_geometric_accuracy'.format(x.lower())] = {'bias': bias, 'stddev': stddev}
    item.properties['card4l:geometric_accuracy_radial_rmse'] = float(meta['prod']['geoCorrAccuracy_rRMSE'])
    #################################################################################
    # Add links
    
    links = [
        {'rel': 'card4l-document',
         'target': meta['prod']['card4l-link'].replace('.pdf', '.docx'),
         'media_type': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
         'title': f'CARD4L Product Family Specification: '
                  f'{meta["prod"]["productName"]} (v{meta["prod"]["card4l-version"]})'},
        {'rel': 'card4l-document',
         'target': meta['prod']['card4l-link'],
         'media_type': 'application/pdf',
         'title': f'CARD4L Product Family Specification: '
                  f'{meta["prod"]["productName"]} (v{meta["prod"]["card4l-version"]})'},
        {"rel": "about",
         "target": meta["prod"]["doi"],
         "title": "Product definition reference.",
         "media_type": None},
        {"rel": "access",
         "target": meta["prod"]["access"],
         "title": "Product data access",
         "media_type": None},
        {"rel": "related",
         "target": meta["prod"]["ancillaryData_KML"],
         "title": "Sentinel-2 Military Grid Reference System (MGRS) tiling "
                  "grid file used as auxiliary data during processing",
         "media_type": None},
        {"rel": "noise-removal",
         "target": meta["prod"]["noiseRemovalAlgorithm"],
         "title": "Reference to the noise removal algorithm details",
         "media_type": None},
        {"rel": "radiometric-terrain-correction",
         "target": meta["prod"]["RTCAlgorithm"],
         "title": "Reference to the Radiometric Terrain Correction algorithm details",
         "media_type": None},
        {"rel": "radiometric-accuracy",
         "target": meta["prod"]["radiometricAccuracyReference"],
         "title": "Reference describing the radiometric uncertainty of the product",
         "media_type": None},
        {"rel": "geometric-correction",
         "target": meta["prod"]["geoCorrAlgorithm"],
         "title": "Reference to the Geometric Correction algorithm details",
         "media_type": None},
        # only one type should be necessary, which is available via meta['prod']['demType']
        # However, the validation fails, because the schema expects both
        {"rel": "elevation-model",
         "target": meta["prod"]["demReference"],
         "title": f"Digital Elevation Model used as auxiliary data during processing: "
                  f"{meta['prod']['demName']}",
         "media_type": None},
        {"rel": "surface-model",
         "target": meta["prod"]["demReference"],
         "title": f"Digital Elevation Model used as auxiliary data during processing: "
                  f"{meta['prod']['demName']}",
         "media_type": None},
        {"rel": "earth-gravitational-model",
         "target": meta["prod"]["demEGMReference"],
         "title": "Reference to the Earth Gravitational Model (EGM) used for geometric correction",
         "media_type": None},
        {"rel": "geometric-accuracy",
         "target": meta["prod"]["geoCorrAccuracyReference"],
         "title": "Reference documenting the estimate of absolute localization error",
         "media_type": None},
        {"rel": "gridding-convention",
         "target": meta["prod"]["griddingConventionURL"],
         "title": "Reference describing the gridding convention used",
         "media_type": None}
    ]
    if meta['prod']['productName-short'] == 'ORB':
        links.append(
            {"rel": "wind-norm-reference",
             "target": meta["prod"]["windNormReferenceModel"],
             "title": "Reference to the model used to create the wind normalisation layer",
             "media_type": None}
        )
    
    for src in list(meta['source'].keys()):
        x = os.path.basename(meta['source'][src]['filename']).split('.')[0]
        src_target = os.path.join('./source', '{}.json'.format(x)).replace('\\', '/')
        links.append(
            {'rel': 'derived_from',
             'target': src_target,
             'media_type': 'application/json',
             'title': 'Source metadata formatted in STAC compliant JSON format'})
    
    for link in links:
        if link['target'] is not None:
            item.add_link(link=pystac.Link(**link))
    #################################################################################
    # Add assets
    assets = assets.copy()
    assets.append(outname.replace('.json', '.xml'))
    
    assets_dict = {'measurement': {},
                   'annotation': {},
                   'metadata': {}}
    for asset in assets:
        relpath = './' + os.path.relpath(asset, target).replace('\\', '/')
        
        size = os.path.getsize(asset)
        hash = compute_hash(asset)
        created = None
        header_size = None
        media_type = pystac.MediaType.COG  # COG, VRT
        byte_order = None
        if asset.endswith('.tif'):
            with Raster(asset) as ras:
                nodata = ras.nodata
            created = datetime.fromtimestamp(os.path.getctime(asset), tz=timezone.utc).isoformat()
            header_size = get_header_size(tif=asset)
            byte_order = ByteOrder.LITTLE_ENDIAN
        
        if 'measurement' in asset:
            key, title = _asset_get_key_title(meta=meta, asset=asset)
            stac_asset = pystac.Asset(href=relpath,
                                      title=title,
                                      media_type=media_type,
                                      roles=['backscatter', 'data'],
                                      extra_fields=None)
            if asset.endswith('.tif'):
                stac_asset.extra_fields = {'created': created,
                                           'card4l:border_pixels': meta['prod']['numBorderPixels']}
                _asset_handle_raster_ext(stac_asset=stac_asset, nodata=nodata)
            file_ext = FileExtension.ext(stac_asset)
            file_ext.apply(byte_order=byte_order, size=size,
                           header_size=header_size, checksum=hash)
            assets_dict['measurement'][key] = stac_asset
        elif 'annotation' in asset:
            key, title = _asset_get_key_title(meta=meta, asset=asset)
            if key == 'np':
                pol = re.search('-[vh]{2}', relpath).group().removeprefix('-')
                asset_key = 'noise-power-{}'.format(pol)
                title = "{} {}".format(title, pol.upper())
            else:
                asset_key = ASSET_MAP[key]['role']
            
            if ASSET_MAP[key]['unit'] is None:
                ASSET_MAP[key]['unit'] = 'unitless'
            
            stac_asset = pystac.Asset(href=relpath,
                                      title=title,
                                      media_type=media_type,
                                      roles=[ASSET_MAP[key]['role'], 'metadata'],
                                      extra_fields=None)
            if key == '-ei.tif':
                stac_asset.extra_fields = {'card4l:ellipsoidal_height': meta['prod']['ellipsoidalHeight']}
            file_ext = FileExtension.ext(stac_asset)
            file_ext.apply(byte_order=byte_order, size=size,
                           header_size=header_size, checksum=hash)
            _asset_handle_raster_ext(stac_asset=stac_asset, nodata=nodata, key=key, meta=meta, asset=asset)
            assets_dict['annotation'][asset_key] = stac_asset
        else:
            stac_asset = pystac.Asset(href=relpath,
                                      title='Metadata in XML format.',
                                      media_type=pystac.MediaType.XML,
                                      roles=['metadata', 'card4l'])
            file_ext = FileExtension.ext(stac_asset)
            file_ext.apply(byte_order=byte_order, size=size,
                           header_size=header_size, checksum=hash)
            assets_dict['metadata']['card4l'] = stac_asset
    
    for category in ['measurement', 'annotation', 'metadata']:
        for key in sorted(assets_dict[category].keys()):
            item.add_asset(key=key, asset=assets_dict[category][key])
    
    if any(x in item.get_assets().keys() for x in ['acquisition-id', 'data-mask']):
        ClassificationExtension.add_to(item)
    FileExtension.add_to(item)
    RasterExtension.add_to(item)
    
    # downgrade some extensions as required by the card4l extension
    exts = list(item.stac_extensions)
    item.stac_extensions = [
        "https://stac-extensions.github.io/file/v2.0.0/schema.json"
        if e.startswith("https://stac-extensions.github.io/file/")
        else "https://stac-extensions.github.io/projection/v1.0.0/schema.json"
        if e.startswith("https://stac-extensions.github.io/projection/")
        else e
        for e in exts
    ]
    
    item.validate()
    item.save_object(dest_href=outname)


def _asset_get_key_title(
        meta: dict[str, dict[str, Any]],
        asset: str
) -> tuple[str, str]:
    """
    Helper function to handle creation of identifying key and title of a given asset.
    
    Parameters
    ----------
    meta:
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
    asset:
        Path to a GeoTIFF or VRT asset.
    
    Returns
    -------
        The key identifying the asset and the generated title of the asset.
    """
    key = None
    title = None
    if 'measurement' in asset:
        title_dict = {'g': 'gamma nought', 's': 'sigma nought',
                      'lin': 'linear', 'log': 'logarithmic'}
        pattern = ('(?P<key>(?P<pol>[vhc]{2})-'
                   '(?P<nought>[gs])-'
                   '(?P<scaling>lin|log)'
                   '(?P<windnorm>-wn|))')
        info = re.search(pattern, asset).groupdict()
        key = info['key']
        
        if re.search('cc-[gs]-lin', key):
            pols = meta['common']['polarisationChannels']
            co = pols.pop(0) if pols[0][0] == pols[0][1] else pols.pop(1)
            cross = pols[0]
            title = 'RGB color composite (' \
                    '{co}-{nought}-lin, ' \
                    '{cross}-{nought}-lin, ' \
                    '{co}-{nought}-lin/{cross}-{nought}-lin)'
            title = title.format(co=co.lower(),
                                 cross=cross.lower(),
                                 nought=info['nought'])
        else:
            if info['windnorm'] == '-wn':
                skeleton = '{pol} {nought} {subtype} wind normalisation ratio, {scale} scaling'
            else:
                skeleton = '{pol} {nought} {subtype} backscatter, {scale} scaling'
            subtype = 'RTC'
            title = skeleton.format(pol=info['pol'].upper(),
                                    nought=title_dict[info['nought']],
                                    subtype=subtype,
                                    scale=title_dict[info['scaling']])
    elif 'annotation' in asset:
        pattern = '(dm|ei|em|lc|ld|li|gs|id|np-[vh]{2}|sg|wm).tif'
        key = re.search(pattern, asset).groups()[0][:2]
        title = ASSET_MAP[key]['title']
    return key, title


def _asset_handle_raster_ext(
        stac_asset: pystac.Asset,
        nodata: float,
        key: str | None = None,
        meta: dict[str, dict[str, Any]] | None = None,
        asset: str | None = None
) -> None:
    """
    Helper function to handle the STAC RasterExtension for a given pystac.Asset
    
    Parameters
    ----------
    stac_asset:
        The pystac.Asset to set the RasterExtension for.
    nodata:
        The asset's nodata value.
    key:
        Key identifying the asset. Only necessary for annotation assets.
    meta: dict
        Metadata dictionary generated by the satellite-specific ARD packages,
        e.g. :func:`s1ard.metadata.extract.meta_dict`.
        Only necessary for annotation assets.
    asset:
        Path to a GeoTIFF or VRT asset. Only necessary for annotation assets.
    """
    raster_ext = RasterExtension.ext(stac_asset)
    if 'measurement' in stac_asset.href:
        raster_ext.bands = [RasterBand.create(nodata=nodata,
                                              data_type=DataType.FLOAT32,
                                              unit='natural')]
    elif 'annotation' in stac_asset.href:
        if key is None or meta is None or asset is None:
            raise ValueError(f'`key`, `meta` and `asset` parameters need to be defined to handle RasterExtension '
                             f'for {stac_asset.href}')
        if key == 'id':
            src_list = [
                os.path.basename(meta['source'][src]['filename']).replace('.SAFE', '').replace('.zip', '')
                for src in list(meta['source'].keys())]
            band = RasterBand.create(nodata=nodata,
                                     data_type=DataType.UINT8,
                                     unit=ASSET_MAP[key]['unit'])
            class_ext = ClassificationExtension.ext(band)
            class_ext.classes = [Classification.create(value=i + 1, description=j) for i, j in enumerate(src_list)]
            raster_ext.bands = [band]
        elif key == 'dm':
            with Raster(asset) as dm_ras:
                band_descr = [dm_ras.raster.GetRasterBand(band).GetDescription() for band in
                              range(1, dm_ras.bands + 1)]
            samples = [x for x in band_descr if x in ASSET_MAP[key]['allowed']]
            bands = []
            for sample in samples:
                band = RasterBand.create(nodata=nodata,
                                         data_type=DataType.UINT8,
                                         unit=ASSET_MAP[key]['unit'])
                class_ext = ClassificationExtension.ext(band, add_if_missing=True)
                class_ext.classes = [Classification.create(value=1, description=sample)]
                bands.append(band)
            raster_ext.bands = bands
        else:
            raster_ext.bands = [RasterBand.create(nodata=nodata,
                                                  data_type=DataType.FLOAT32,
                                                  unit=ASSET_MAP[key]['unit'])]
    if key == 'em':
        raster_ext.bands[0].spatial_resolution = int(meta['prod']['demGSD'].split()[0])



import os
import re
import inspect
from dateutil.parser import parse as dateparse
from datetime import datetime, timedelta
from spatialist.vector import Vector, crsConvert, wkt2vector
import asf_search as asf
from pyroSAR.drivers import ID
from cesard.ancillary import date_to_utc, combine_polygons
from cesard.tile_extraction import aoi_from_tile, tile_from_aoi
from typing import Any, Literal
import logging

log = logging.getLogger('cesard')

SENSOR = Literal["S1A", "S1B", "S1C", "S1D"]
PRODUCT = Literal["GRD", "SLC"]
ACQUISITION_MODE = Literal["IW", "EW", "SM"]


class ASF(ID):
    """
    Simple SAR metadata handler for scenes in the ASF archive. The interface
    is consistent with the driver classes in :mod:`pyroSAR.drivers` but does
    not implement the full functionality due to limited content of the CMR
    metadata catalog. Registered attributes:
    
    - acquisition_mode
    - coordinates
    - frameNumber
    - orbit
    - orbitNumber_abs
    - orbitNumber_rel
    - polarizations
    - product
    - projection
    - sensor
    - start
    - stop
    """
    
    def __init__(
            self,
            meta: dict[str, Any]
    ):
        self.scene = meta['properties']['url']
        self._meta = meta
        self.meta = self.scanMetadata()
        super(ASF, self).__init__(self.meta)
    
    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, ASF):
            return NotImplemented
        return self.outname_base() < other.outname_base()
    
    def scanMetadata(self) -> dict[str, int | float | str | datetime | list[tuple[float, float]]]:
        meta = dict()
        meta['acquisition_mode'] = self._meta['properties']['beamModeType']
        meta['coordinates'] = [tuple(x) for x in self._meta['geometry']['coordinates'][0]]
        fname = os.path.splitext(self._meta['properties']['fileName'])[0]
        meta['frameNumber'] = fname[-11:-5]
        meta['orbit'] = self._meta['properties']['flightDirection'][0]
        meta['orbitNumber_abs'] = self._meta['properties']['orbit']
        meta['orbitNumber_rel'] = self._meta['properties']['pathNumber']
        meta['polarizations'] = self._meta['properties']['polarization'].split('+')
        product = self._meta['properties']['processingLevel']
        meta['product'] = re.search('GRD|SLC|SM|OCN', product).group()
        meta['projection'] = crsConvert(4326, 'wkt')
        meta['sensor'] = self._meta['properties']['platform'].replace('entinel-', '')
        start = self._meta['properties']['startTime']
        stop = self._meta['properties']['stopTime']
        pattern = '%Y%m%dT%H%M%S'
        meta['start'] = dateparse(start).strftime(pattern)
        meta['stop'] = dateparse(stop).strftime(pattern)
        meta['spacing'] = None
        meta['samples'] = None
        meta['lines'] = None
        meta['cycleNumber'] = None
        meta['sliceNumber'] = None
        meta['totalSlices'] = None
        return meta


class ASFArchive(object):
    """
    Search for scenes in the Alaska Satellite Facility (ASF) catalog.
    """
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        return
    
    @staticmethod
    def select(
            sensor: SENSOR | None = None,
            product: PRODUCT | None = None,
            acquisition_mode: ACQUISITION_MODE | None = None,
            mindate: str | datetime | None = None,
            maxdate: str | datetime | None = None,
            vectorobject: Vector | None = None,
            date_strict: bool = True,
            return_value: str | list[str] = 'scene'
    ) -> list[str | tuple[str] | ASF]:
        """
        Select scenes from the ASF catalog. This is a simple wrapper around the function
        :func:`~cesard.search.asf_select` to be consistent with the interfaces of the
        other search classes.

        Parameters
        ----------
        sensor:
            the satellite
        product:
            the product type
        acquisition_mode:
            the satellite acquisition mode
        mindate:
            the minimum acquisition date; timezone-unaware dates are interpreted as UTC.
        maxdate:
            the maximum acquisition date; timezone-unaware dates are interpreted as UTC.
        vectorobject:
            a geometry with which the scenes need to overlap. The object may only contain one feature.
        date_strict:
            treat dates as strict limits or also allow flexible limits to incorporate scenes
            whose acquisition period overlaps with the defined limit?
            
            - strict: start >= mindate & stop <= maxdate
            - not strict: stop >= mindate & start <= maxdate
        return_value:
            the metadata return value; see :func:`~cesard.search.asf_select` for details
        
        Returns
        -------
            the scene metadata attributes as specified with `return_value`;
            see :func:`~cesard.search.asf_select` for details
        
        See Also
        --------
        asf_select
        """
        return asf_select(sensor, product, acquisition_mode, mindate, maxdate, vectorobject,
                          return_value=return_value, date_strict=date_strict)


def asf_select(
        sensor: SENSOR | None = None,
        product: PRODUCT | None = None,
        acquisition_mode: ACQUISITION_MODE | None = None,
        mindate: str | datetime | None = None,
        maxdate: str | datetime | None = None,
        vectorobject: Vector | None = None,
        date_strict: bool = True,
        return_value: str | list[str] = 'scene'
) -> list[str | tuple[str] | ASF]:
    """
    Search scenes in the Alaska Satellite Facility (ASF) data catalog.
    This is a simple interface to the
    `asf_search <https://github.com/asfadmin/Discovery-asf_search>`_ package.
    
    Parameters
    ----------
    sensor:
        the satellite
    product:
        the product type
    acquisition_mode:
        the satellite acquisition mode
    mindate:
        the minimum acquisition date; timezone-unaware dates are interpreted as UTC.
    maxdate:
        the maximum acquisition date; timezone-unaware dates are interpreted as UTC.
    vectorobject:
        a geometry with which the scenes need to overlap. The object may only contain one feature.
    date_strict:
        treat dates as strict limits or also allow flexible limits to incorporate scenes
        whose acquisition period overlaps with the defined limit?
        
        - strict: start >= mindate & stop <= maxdate
        - not strict: stop >= mindate & start <= maxdate
    
    return_value:
        the query return value(s). Options:
        
        - acquisition_mode: the sensor's acquisition mode
        - frameNumber: the frame or datatake number
        - geometry_wkb: the scene's footprint geometry formatted as WKB
        - geometry_wkt: the scene's footprint geometry formatted as WKT
        - mindate: the acquisition start datetime in UTC formatted as YYYYmmddTHHMMSS
        - maxdate: the acquisition end datetime in UTC formatted as YYYYmmddTHHMMSS
        - product: the product type
        - scene: the scene's storage location path (default)
        - sensor: the satellite platform

    Returns
    -------
        the scene metadata attributes as specified with `return_value`; the return type
        is a list of strings, tuples, or :class:`~cesard.search.ASF` objects depending on
        whether `return_type` is of type string, list or :class:`~cesard.search.ASF`.
    
    """
    if isinstance(return_value, str):
        return_values = [return_value]
    else:
        return_values = return_value
    
    if product == 'GRD':
        processing_level = ['GRD_HD', 'GRD_MD', 'GRD_MS', 'GRD_HS', 'GRD_FD']
    else:
        processing_level = product
    if acquisition_mode == 'SM':
        beam_mode = ['S1', 'S2', 'S3', 'S4', 'S5', 'S6']
    else:
        beam_mode = acquisition_mode
    if vectorobject is not None:
        if vectorobject.nfeatures > 1:
            raise RuntimeError("'vectorobject' contains more than one feature.")
        with vectorobject.clone() as geom:
            geom.reproject(4326)
            geometry = geom.convert2wkt(set3D=False)[0]
    else:
        geometry = None
    
    start = date_to_utc(mindate, as_datetime=True)
    stop = date_to_utc(maxdate, as_datetime=True)
    
    lookup_platform = {'S1A': 'Sentinel-1A',
                       'S1B': 'Sentinel-1B',
                       'S1C': 'Sentinel-1C',
                       'S1D': 'Sentinel-1D'}
    platform = lookup_platform[sensor] if sensor is not None else None
    
    result = asf.search(platform=platform,
                        processingLevel=processing_level,
                        beamMode=beam_mode,
                        start=start,
                        end=stop,
                        intersectsWith=geometry).geojson()
    features = result['features']
    
    def date_extract(item, key):
        return date_to_utc(date=item['properties'][key], as_datetime=True)
    
    if date_strict:
        features = [x for x in features
                    if start <= date_extract(x, 'startTime')
                    and date_extract(x, 'stopTime') <= stop]
    
    features = sorted([ASF(x) for x in features])
    
    out = []
    for item in features:
        values = []
        for key in return_values:
            if key == 'ASF':
                values.append(item)
            elif key == 'mindate':
                values.append(getattr(item, 'start'))
            elif key == 'maxdate':
                values.append(getattr(item, 'stop'))
            elif key == 'geometry_wkb':
                with item.geometry() as vec:
                    value = vec.to_geopandas().to_wkb()['geometry'][0]
                    values.append(value)
            elif key == 'geometry_wkt':
                with item.geometry() as vec:
                    value = vec.to_geopandas().to_wkt()['geometry'][0]
                    values.append(value)
            elif hasattr(item, key):
                values.append(getattr(item, key))
            else:
                raise ValueError(f'invalid return value: {key}')
        if len(return_values) == 1:
            out.append(values[0])
        else:
            out.append(tuple(values))
    return out


def scene_select(
        archive: Any,
        aoi_tiles: list[str] | None = None,
        aoi_geometry: str | None = None,
        return_value: str | list[str] = 'scene',
        **kwargs: Any
) -> tuple[list[str | tuple[str]], list[str]]:
    """
    Central scene search utility. Selects scenes from a database and returns their file names
    together with the MGRS tile names for which to process ARD products.
    The list of MGRS tile names is either identical to the list provided with `aoi_tiles`,
    the list of all tiles overlapping with `aoi_geometry` or `vectorobject` (via `kwargs`),
    or the list of all tiles overlapping with an initial scene search result if no geometry
    has been defined via `aoi_tiles` or `aoi_geometry`. In the latter (most complex) case,
    the search procedure is as follows:
    
     - perform a first search matching all other search parameters
     - derive all MGRS tile geometries overlapping with the selection
     - derive the minimum and maximum acquisition times of the selection as search parameters
       `mindate` and `maxdate`
     - extend the `mindate` and `maxdate` search parameters by one minute
     - perform a second search with the extended time range and the derived MGRS tile geometries
     - filter the search result to scenes overlapping with the initial time range (if defined
       via `mindate` or `maxdate`)
    
    As consequence, if one defines the search parameters to only return one scene, the neighboring
    acquisitions will also be returned. This is because the scene overlaps with a set of MGRS
    tiles of which many or all will also overlap with these neighboring acquisitions. To ensure
    full coverage of all MGRS tiles, the neighbors of the scene in focus have to be processed too.
    
    This function has three ways to define search geometries. In order of priority overriding others:
    `aoi_tiles` > `aoi_geometry` > `vectorobject` (via `kwargs`). In the latter two cases, the search
    geometry is extended to the common footprint of all MGRS tiles overlapping with the initial geometry
    to ensure full coverage of all tiles.
    
    Parameters
    ----------
    archive:
        an open scene archive connection. pyroSAR.drivers.Archive or s1ard.search.STACArchive
        or s1ard.search.STACParquetArchive or ASFArchive.
    aoi_tiles:
        a list of MGRS tile names for spatial search
    aoi_geometry:
        the name of a vector geometry file for spatial search
    return_value:
        the query return value(s). Default 'scene': return the scene's storage location path.
        See the documentation of `archive.select` for options.
    kwargs:
        further search arguments passed to the `select` method of `archive`.
        The `date_strict` argument has no effect. Whether an ARD product is strictly in the defined
        time range cannot be determined by this function, and it thus has to add a time buffer.
        When `date_strict=True`, more scenes will be filtered out in the last step described above.

    Returns
    -------
        a tuple containing
    
        - the list of return values; single value:string, multiple values: tuple
        - the list of MGRS tiles
    
    """
    args = kwargs.copy()
    if 'mindate' in args.keys():
        args['mindate'] = date_to_utc(args['mindate'], as_datetime=True)
        mindate_init = args['mindate']
    else:
        mindate_init = None
    if 'maxdate' in args.keys():
        args['maxdate'] = date_to_utc(args['maxdate'], as_datetime=True)
        maxdate_init = args['maxdate']
    else:
        maxdate_init = None
    for key in ['acquisition_mode']:
        if key not in args.keys():
            args[key] = None
    
    if args['acquisition_mode'] == 'SM':
        args['acquisition_mode'] = ('S1', 'S2', 'S3', 'S4', 'S5', 'S6')
    
    signature = inspect.signature(archive.select)
    if 'return_value' not in signature.parameters:
        raise RuntimeError("the 'select' method of 'archive' does not take "
                           "a 'return_value' parameter")
    
    return_values = return_value if isinstance(return_value, list) else [return_value]
    if isinstance(archive, ASFArchive):
        args['return_value'] = 'ASF'
    else:
        args['return_value'] = return_value
    
    vec = None
    if aoi_tiles is not None:
        log.debug("reading geometries of 'aoi_tiles'")
        vec = aoi_from_tile(tile=aoi_tiles)
    elif aoi_geometry is not None:
        log.debug("extracting tiles overlapping with 'aoi_geometry'")
        with Vector(aoi_geometry) as geom:
            vec = tile_from_aoi(vector=geom,
                                return_geometries=True)
    elif 'vectorobject' in args.keys() and args['vectorobject'] is not None:
        log.debug("extracting tiles overlapping with 'vectorobject'")
        vec = tile_from_aoi(vector=args['vectorobject'],
                            return_geometries=True)
    if vec is not None:
        if not isinstance(vec, list):
            vec = [vec]
        
        if aoi_tiles is None:
            aoi_tiles = [x.mgrs for x in vec]
        log.debug(f"got {len(aoi_tiles)} tiles")
    
    # derive geometries and tiles from scene footprints
    if vec is None:
        log.debug("performing initial scene search without geometry constraint")
        args['return_value'] = ['mindate', 'maxdate', 'geometry_wkt']
        selection_tmp = archive.select(**args)
        log.debug(f'got {len(selection_tmp)} scenes')
        mindates, maxdates, geometries_init = zip(*selection_tmp)
        # The geometry of scenes crossing the antimeridian is stored as multipolygon.
        # Since the processor is currently not able to process these scenes, they are
        # removed in this step.
        geometries = [x for x in geometries_init if x.startswith('POLYGON')]
        if len(geometries) < len(geometries_init):
            log.debug(f'removed {len(geometries_init) - len(geometries)} '
                      f'scenes crossing the antimeridian')
        del selection_tmp
        
        log.debug(f"loading geometries")
        scenes_geom = [wkt2vector(x, srs=4326) for x in geometries]
        # select all tiles overlapping with the scenes for further processing
        log.debug("extracting all tiles overlapping with initial scene selection")
        vec = tile_from_aoi(vector=scenes_geom,
                            return_geometries=True)
        if not isinstance(vec, list):
            vec = [vec]
        aoi_tiles = [x.mgrs for x in vec]
        log.debug(f"got {len(aoi_tiles)} tiles")
        del scenes_geom
        
        args['mindate'] = min([date_to_utc(x, as_datetime=True) for x in mindates])
        args['maxdate'] = max([date_to_utc(x, as_datetime=True) for x in maxdates])
        del mindates, maxdates, geometries
    
    # extend the time range to fully cover all tiles
    # (one additional scene needed before and after each data take group)
    if 'mindate' in args.keys():
        args['mindate'] -= timedelta(minutes=1)
    if 'maxdate' in args.keys():
        args['maxdate'] += timedelta(minutes=1)
    
    args['return_value'] = return_values.copy()
    for key in ['mindate', 'maxdate']:
        if key not in args['return_value']:
            args['return_value'].append(key)
    
    log.debug("performing main scene search")
    with combine_polygons(vec, multipolygon=True) as combined:
        args['vectorobject'] = combined
        selection = archive.select(**args)
    del vec
    
    # reduce the selection to the time range defined by the user
    if mindate_init is not None or maxdate_init is not None:
        i = 0
        while i < len(selection):
            values = dict(zip(args['return_value'], selection[i]))
            start = date_to_utc(values['mindate'], as_datetime=True)
            stop = date_to_utc(values['maxdate'], as_datetime=True)
            delete = False
            if mindate_init is not None and stop < mindate_init:
                delete = True
            if maxdate_init is not None and start > maxdate_init:
                delete = True
            if delete:
                del selection[i]
            else:
                i += 1
    
    # sort the return values by the scene's basename
    rv_scene_key = args['return_value'].index('scene')
    selection = sorted(selection, key=lambda k: os.path.basename(k[rv_scene_key]))
    
    # reduce the return values to those defined by the user
    indices = [i for i, key in enumerate(args['return_value']) if key in return_values]
    if len(indices) == 1:
        selection = [scene[indices[0]] for scene in selection]
    else:
        selection = [tuple(scene[i] for i in indices) for scene in selection]
    log.debug(f"got {len(selection)} scenes")
    return selection, aoi_tiles



import os
import re
from spatialist import Raster, Vector
from spatialist.envi import HDRobject
from spatialist.ancillary import finder
from pyroSAR import identify, identify_many
from pyroSAR.examine import ExamineSnap
from pyroSAR.snap.auxil import gpt, parse_recipe, parse_node, \
    mli_parametrize, geo_parametrize, \
    sub_parametrize, erode_edges
import cesard
import logging

log = logging.getLogger('cesard')


def find_datasets(
        scene: str,
        outdir: str,
        epsg: int
) -> dict[str, str] | None:
    """
    Find processed datasets for a SAR scene.

    Parameters
    ----------
    scene:
        the file name of the SAR scene
    outdir:
        the output directory in which to search for results
    epsg:
        the EPSG code defining the output projection of the processed products.

    Returns
    -------
        Either None if no datasets were found or a dictionary with the
        following keys and values pointing to the file names
        (polarization-specific keys depending on product availability):

         - hh-g-lin: gamma nought RTC backscatter HH polarization
         - hv-g-lin: gamma nought RTC backscatter HV polarization
         - vh-g-lin: gamma nought RTC backscatter VH polarization
         - vv-g-lin: gamma nought RTC backscatter VV polarization
         - hh-s-lin: sigma nought ellipsoidal backscatter HH polarization
         - hv-s-lin: sigma nought ellipsoidal backscatter HV polarization
         - vh-s-lin: sigma nought ellipsoidal backscatter VH polarization
         - vv-s-lin: sigma nought ellipsoidal backscatter VV polarization
         - dm: layover-shadow data mask
         - ei: ellipsoidal incident angle
         - gs: gamma-sigma ratio
         - lc: local contributing area (aka scattering area)
         - ld: range look direction angle
         - li: local incident angle
         - sg: sigma-gamma ratio
         - np-hh: NESZ HH polarization
         - np-hv: NESZ HV polarization
         - np-vh: NESZ VH polarization
         - np-vv: NESZ VV polarization
    """
    basename = os.path.splitext(os.path.basename(scene))[0]
    scenedir = os.path.join(outdir, basename)
    subdir = os.path.join(scenedir, basename + f'_geo_{epsg}.data')
    if not os.path.isdir(subdir):
        return
    lookup = {'dm': r'layoverShadowMask\.img$',
              'ei': r'incidenceAngleFromEllipsoid\.img$',
              'gs': r'gammaSigmaRatio_[VH]{2}\.img$',
              'lc': r'simulatedImage_[VH]{2}\.img$',
              'ld': r'lookDirection_[VH]{2}\.img$',
              'li': r'localIncidenceAngle\.img$',
              'sg': r'sigmaGammaRatio_[VH]{2}\.img$'}
    out = {}
    for key, pattern in lookup.items():
        match = finder(target=subdir, matchlist=[pattern], regex=True)
        if len(match) > 0:
            out[key] = match[0]
    pattern = r'(?P<bsc>Gamma0|Sigma0)_(?P<pol>[VH]{2})\.img$'
    backscatter = finder(target=subdir, matchlist=[pattern], regex=True)
    for item in backscatter:
        pol = re.search(pattern, item).group('pol').lower()
        bsc = re.search(pattern, item).group('bsc')[0].lower()
        out[f'{pol}-{bsc}-lin'] = item
    pattern = r'NESZ_(?P<pol>[VH]{2})\.img$'
    nesz = finder(target=subdir, matchlist=[pattern], regex=True)
    for item in nesz:
        pol = re.search(pattern, item).group('pol')
        out[f'np-{pol.lower()}'] = item
    if len(out) > 0:
        return out


def lsm_encoding() -> dict[str, int]:
    """
    Get the encoding of the layover shadow mask.
    """
    return {
        'not layover, not shadow': 0,
        'layover': 1,
        'shadow': 2,
        'layover in shadow': 3,
        'nodata': 255  # dummy value
    }


def version_dict() -> dict[str, str]:
    """
    Get processor software version information.

    Returns
    -------
        a dictionary with software components as keys and their versions as values.
    """
    try:
        snap_config = ExamineSnap()
        core = snap_config.get_version('core')
        microwavetbx = snap_config.get_version('microwavetbx')
        snap_core = f"{core['version']} | {core['date']}"
        snap_microwavetbx = f"{microwavetbx['version']} | {microwavetbx['date']}"
    except RuntimeError:
        snap_core = 'unknown'
        snap_microwavetbx = 'unknown'
    return ({'cesard': cesard.__version__,
             'snap_core': snap_core,
             'snap_microwavetbx': snap_microwavetbx})


###############################################################################

def geo(
        *src: str | None,
        dst: str,
        workflow: str,
        spacing: int | float,
        crs: int | float,
        geometry: dict[str, int | float] | Vector | str | None = None,
        buffer: int | float = 0.01,
        export_extra: list[str] | None = None,
        standard_grid_origin_x: int | float = 0,
        standard_grid_origin_y: int | float = 0,
        dem: str,
        dem_resampling_method: str = 'BILINEAR_INTERPOLATION',
        img_resampling_method: str = 'BILINEAR_INTERPOLATION',
        gpt_args: list[str] | None = None,
        **bands
) -> None:
    """
    Range-Doppler geocoding.

    Parameters
    ----------
    src:
        variable number of input scene file names
    dst:
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow:
        the target XML workflow file name
    spacing:
        the target pixel spacing in meters
    crs:
        the target coordinate reference system
    geometry:
        a vector geometry to limit the target product's extent
    buffer:
        an additional buffer in degrees to add around `geometry`
    export_extra:
        a list of ancillary layers to write. Supported options:

         - DEM
         - incidenceAngleFromEllipsoid
         - layoverShadowMask
         - localIncidenceAngle
         - projectedLocalIncidenceAngle
    standard_grid_origin_x:
        the X coordinate for pixel alignment
    standard_grid_origin_y:
        the Y coordinate for pixel alignment
    dem:
        the DEM file
    dem_resampling_method:
        the DEM resampling method
    img_resampling_method:
        the SAR image resampling method
    gpt_args:
        a list of additional arguments to be passed to the gpt call

        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing
    bands:
        band ids for the input scenes in `src` as lists with keys bands<index>,
        e.g., ``bands1=['NESZ_VV'], bands2=['Gamma0_VV'], ...``

    See Also
    --------
    pyroSAR.snap.auxil.sub_parametrize
    pyroSAR.snap.auxil.geo_parametrize
    """
    wf = parse_recipe('blank')
    ############################################
    scenes = identify_many(list(filter(None, src)))
    read_ids = []
    for i, scene in enumerate(scenes):
        read = parse_node('Read')
        read.parameters['file'] = scene.scene
        if f'bands{i}' in bands.keys():
            read.parameters['useAdvancedOptions'] = True
            read.parameters['sourceBands'] = bands[f'bands{i}']
        wf.insert_node(read)
        read_ids.append(read.id)
    ############################################
    if len(scenes) > 1:
        merge = parse_node('BandMerge')
        wf.insert_node(merge, before=read_ids)
        last = merge
    else:
        last = wf['Read']
    ############################################
    if geometry is not None:
        sub = sub_parametrize(scene=scenes[0], geometry=geometry, buffer=buffer)
        wf.insert_node(sub, before=last.id)
        last = sub
    ############################################
    tc = geo_parametrize(spacing=spacing, t_srs=crs,
                         export_extra=export_extra,
                         alignToStandardGrid=True,
                         externalDEMFile=dem,
                         externalDEMApplyEGM=False,
                         standardGridOriginX=standard_grid_origin_x,
                         standardGridOriginY=standard_grid_origin_y,
                         standardGridAreaOrPoint='area',
                         demResamplingMethod=dem_resampling_method,
                         imgResamplingMethod=img_resampling_method)
    wf.insert_node(tc, before=last.id)
    ############################################
    write = parse_node('Write')
    wf.insert_node(write, before=tc.id)
    write.parameters['file'] = dst
    write.parameters['formatName'] = 'BEAM-DIMAP'
    ############################################
    wf.write(workflow)
    gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
        gpt_args=gpt_args)


def gsr(
        src: str,
        dst: str,
        workflow: str,
        src_sigma: str | None = None,
        gpt_args: list[str] | None = None
) -> None:
    """
    Gamma-sigma ratio computation for either ellipsoidal or RTC sigma nought.

    Parameters
    ----------
    src:
        the file name of the source scene. Both gamma and sigma bands are expected unless `src_sigma` is defined.
    dst:
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow:
        the output SNAP XML workflow filename.
    src_sigma:
        the optional file name of a second source product from which to read the sigma band.
    gpt_args:
        a list of additional arguments to be passed to the gpt call

        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing

    """
    scene = identify(src)
    pol = scene.polarizations[0]
    wf = parse_recipe('blank')
    ############################################
    read = parse_node('Read')
    read.parameters['file'] = scene.scene
    wf.insert_node(read)
    last = read
    ############################################
    if src_sigma is not None:
        read.parameters['sourceBands'] = f'Gamma_{pol}'
        read2 = parse_node('Read')
        read2.parameters['file'] = src_sigma
        read2.parameters['sourceBands'] = f'Sigma_{pol}'
        wf.insert_node(read2)
        ########################################
        merge = parse_node('BandMerge')
        wf.insert_node(merge, before=[read.id, read2.id])
        last = merge
    ############################################
    math = parse_node('BandMaths')
    wf.insert_node(math, before=last.id)
    ratio = 'gammaSigmaRatio'
    expression = f'Sigma0_{pol} / Gamma0_{pol}'
    
    math.parameters.clear_variables()
    exp = math.parameters['targetBands'][0]
    exp['name'] = ratio
    exp['type'] = 'float32'
    exp['expression'] = expression
    exp['noDataValue'] = 0.0
    ############################################
    write = parse_node('Write')
    wf.insert_node(write, before=math.id)
    write.parameters['file'] = dst
    write.parameters['formatName'] = 'BEAM-DIMAP'
    ############################################
    wf.write(workflow)
    gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
        gpt_args=gpt_args)


def mli(
        src: str,
        dst: str,
        workflow: str,
        spacing: int | float = None,
        rlks: int | None = None,
        azlks: int | None = None,
        gpt_args: list[str] | None = None
) -> None:
    """
    Multi-looking.

    Parameters
    ----------
    src:
        the file name of the source scene
    dst:
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow:
        the output SNAP XML workflow filename.
    spacing:
        the target pixel spacing for automatic determination of looks
        using function :func:`pyroSAR.ancillary.multilook_factors`.
        Overridden by arguments `rlks` and `azlks` if they are not None.
    rlks:
        the number of range looks.
    azlks:
        the number of azimuth looks.
    gpt_args:
        a list of additional arguments to be passed to the gpt call

        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing

    See Also
    --------
    pyroSAR.snap.auxil.mli_parametrize
    pyroSAR.ancillary.multilook_factors
    """
    scene = identify(src)
    wf = parse_recipe('blank')
    ############################################
    read = parse_node('Read')
    read.parameters['file'] = scene.scene
    wf.insert_node(read)
    ############################################
    ml = mli_parametrize(scene=scene, spacing=spacing, rlks=rlks, azlks=azlks)
    if ml is not None:
        log.info('multi-looking')
        wf.insert_node(ml, before=read.id)
        ############################################
        write = parse_node('Write')
        wf.insert_node(write, before=ml.id)
        write.parameters['file'] = dst
        write.parameters['formatName'] = 'BEAM-DIMAP'
        ############################################
        wf.write(workflow)
        gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
            gpt_args=gpt_args)


def postprocess(
        src: str,
        clean_edges: bool = True,
        clean_edges_pixels: int = 4):
    """
    Performs edge cleaning and sets the nodata value in the output ENVI HDR files.

    Parameters
    ----------
    src:
        the file name of the source scene. Format is BEAM-DIMAP.
    clean_edges:
        perform edge cleaning?
    clean_edges_pixels:
        the number of pixels to erode during edge cleaning.

    Returns
    -------

    """
    if clean_edges:
        erode_edges(src=src, only_boundary=True, pixels=clean_edges_pixels)
    datadir = src.replace('.dim', '.data')
    hdrfiles = finder(target=datadir, matchlist=['*.hdr'])
    for hdrfile in hdrfiles:
        if not 'layoverShadowMask' in hdrfile:
            with HDRobject(hdrfile) as hdr:
                hdr.data_ignore_value = 0
                hdr.write(hdrfile)


def rtc(
        src: str,
        dst: str,
        workflow: str,
        dem: str,
        dem_resampling_method: str = 'BILINEAR_INTERPOLATION',
        sigma0: bool = True,
        scattering_area: bool = True,
        dem_oversampling_multiple: int = 2,
        gpt_args: list[str] | None = None
) -> None:
    """
    Radiometric Terrain Flattening.
    
    Parameters
    ----------
    src:
        the file name of the source scene
    dst:
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow:
        the output SNAP XML workflow filename.
    dem:
        the input DEM file name.
    dem_resampling_method:
        the DEM resampling method.
    sigma0:
        output sigma0 RTC backscatter?
    scattering_area:
        output scattering area image?
    dem_oversampling_multiple:
        a factor to multiply the DEM oversampling factor computed by SNAP.
        The SNAP default of 1 has been found to be insufficient with stripe
        artifacts remaining in the image.
    gpt_args:
        a list of additional arguments to be passed to the gpt call
        
        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing
    """
    scene = identify(src)
    wf = parse_recipe('blank')
    ############################################
    read = parse_node('Read')
    read.parameters['file'] = scene.scene
    wf.insert_node(read)
    ############################################
    tf = parse_node('Terrain-Flattening')
    polarizations = scene.polarizations
    bands = ['Beta0_{}'.format(pol) for pol in polarizations]
    wf.insert_node(tf, before=read.id)
    tf.parameters['sourceBands'] = bands
    if 'reGridMethod' in tf.parameters.keys():
        tf.parameters['reGridMethod'] = False
    tf.parameters['outputSigma0'] = sigma0
    tf.parameters['outputSimulatedImage'] = scattering_area
    tf.parameters['demName'] = 'External DEM'
    tf.parameters['externalDEMFile'] = dem
    tf.parameters['externalDEMApplyEGM'] = False
    with Raster(dem) as ras:
        tf.parameters['externalDEMNoDataValue'] = ras.nodata
    tf.parameters['demResamplingMethod'] = dem_resampling_method
    tf.parameters['oversamplingMultiple'] = dem_oversampling_multiple
    last = tf
    ############################################
    write = parse_node('Write')
    wf.insert_node(write, before=last.id)
    write.parameters['file'] = dst
    write.parameters['formatName'] = 'BEAM-DIMAP'
    ############################################
    wf.write(workflow)
    gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
        gpt_args=gpt_args)


def sgr(
        src: str,
        dst: str,
        workflow: str,
        src_gamma: str | None = None,
        gpt_args: list[str] | None = None):
    """
    Sigma-gamma ratio computation.

    Parameters
    ----------
    src: str
        the file name of the source scene. Both sigma and gamma bands are expected unless `src_gamma` is defined.
    dst: str
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow: str
        the output SNAP XML workflow filename.
    src_gamma: str or None
        the optional file name of a second source product from which to read the gamma band.
    gpt_args: list[str] or None
        a list of additional arguments to be passed to the gpt call
        
        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing
    
    Returns
    -------

    """
    scene = identify(src)
    pol = scene.polarizations[0]
    wf = parse_recipe('blank')
    ############################################
    read = parse_node('Read')
    read.parameters['file'] = scene.scene
    wf.insert_node(read)
    last = read
    ############################################
    if src_gamma is not None:
        read.parameters['sourceBands'] = f'Sigma_{pol}'
        read2 = parse_node('Read')
        read2.parameters['file'] = src_gamma
        read2.parameters['sourceBands'] = f'Gamma_{pol}'
        wf.insert_node(read2)
        ########################################
        merge = parse_node('BandMerge')
        wf.insert_node(merge, before=[read.id, read2.id])
        last = merge
    ############################################
    math = parse_node('BandMaths')
    wf.insert_node(math, before=last.id)
    ratio = 'sigmaGammaRatio'
    expression = f'Gamma0_{pol} / Sigma0_{pol}'
    
    math.parameters.clear_variables()
    exp = math.parameters['targetBands'][0]
    exp['name'] = ratio
    exp['type'] = 'float32'
    exp['expression'] = expression
    exp['noDataValue'] = 0.0
    ############################################
    write = parse_node('Write')
    wf.insert_node(write, before=math.id)
    write.parameters['file'] = dst
    write.parameters['formatName'] = 'BEAM-DIMAP'
    ############################################
    wf.write(workflow)
    gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
        gpt_args=gpt_args)



import re
import json
import numpy as np
from datetime import datetime
from spatialist import Raster
from spatialist.auxil import crsConvert
from spatialist.vector import Vector
from osgeo import gdal, ogr
from typing import Any

gdal.UseExceptions()


def vec_from_srccoords(
        coord_list: list[list[tuple[float, float]]],
        crs: int | str,
        layername: str = 'polygon'
) -> Vector:
    """
    Creates a single :class:`~spatialist.vector.Vector` object from a list
    of footprint coordinates of source scenes.
    
    Parameters
    ----------
    coord_list:
        List containing for each source scene a list of coordinate pairs as
        retrieved from the metadata stored in an :class:`~pyroSAR.drivers.ID`
        object.
    crs:
        the coordinate reference system of the provided coordinates.
    layername:
        the layer name of the output vector object
    
    Returns
    -------
        the vector object
    """
    srs = crsConvert(crs, 'osr')
    pts = ogr.Geometry(ogr.wkbMultiPoint)
    for footprint in coord_list:
        for lon, lat in footprint:
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(lon, lat)
            pts.AddGeometry(point)
    geom = pts.ConvexHull()
    vec = Vector(driver='Memory')
    vec.addlayer(layername, srs, geom.GetGeometryType())
    vec.addfeature(geom)
    point = None
    pts = None
    geom = None
    return vec


def geometry_from_vec(
        vectorobject: Vector
) -> dict[str, Any]:
    """
    Get geometry information for usage in STAC and XML metadata from a :class:`spatialist.vector.Vector` object.
    
    Parameters
    ----------
    vectorobject:
        The vector object to extract geometry information from.
    
    Returns
    -------
        A dictionary containing the geometry information extracted from the vector object.
    """
    out = {}
    vec = vectorobject
    
    # For STAC metadata
    if vec.getProjection(type='epsg') != 4326:
        ext = vec.extent
        out['bbox_native'] = [ext['xmin'], ext['ymin'], ext['xmax'], ext['ymax']]
        vec.reproject(4326)
    feat = vec.getfeatures()[0]
    geom = feat.GetGeometryRef()
    out['geometry'] = json.loads(geom.ExportToJson())
    ext = vec.extent
    out['bbox'] = [ext['xmin'], ext['ymin'], ext['xmax'], ext['ymax']]
    
    # For XML metadata
    c_x = (ext['xmax'] + ext['xmin']) / 2
    c_y = (ext['ymax'] + ext['ymin']) / 2
    out['center'] = '{} {}'.format(c_y, c_x)
    wkt = geom.ExportToWkt().removeprefix('POLYGON ((').removesuffix('))')
    wkt_list = ['{} {}'.format(x[1], x[0]) for x in [y.split(' ') for y in wkt.split(',')]]
    out['envelope'] = ' '.join(wkt_list)
    
    return out


def calc_enl(
        tif: str,
        block_size: int = 30,
        return_arr: bool = False,
        decimals: int = 2
) -> float | np.ndarray | None:
    """
    Calculate the Equivalent Number of Looks (ENL) for a linear-scaled backscatter
    measurement GeoTIFF file. The calculation is performed block-wise for the
    entire image and by default the median ENL value is returned.
    
    Parameters
    ----------
    tif:
        The path to a linear-scaled backscatter measurement GeoTIFF file.
    block_size:
        The block size to use for the calculation. Remainder pixels are discarded,
         if the array dimensions are not evenly divisible by the block size.
         Default is 30, which calculates ENL for 30x30 pixel blocks.
    return_arr:
        If True, the calculated ENL array is returned. Default is False.
    decimals:
        Number of decimal places to round the calculated ENL value to. Default is 2.
    
    Raises
    ------
    RuntimeError
        if the input array contains only NaN values
    
    Returns
    -------
        If `return_enl_arr=True`, an array of ENL values is returned. Otherwise,
        the median ENL value is returned. If the ENL array contains only NaN and
        `return_enl_arr=False`, the return value is `None`.
    
    References
    ----------
    :cite:`anfinsen.etal_2009`
    """
    with Raster(tif) as ras:
        arr = ras.array()
    arr[np.isinf(arr)] = np.nan
    
    if len(arr[~np.isnan(arr)]) == 0:
        raise RuntimeError('cannot compute ENL for an empty array')
    
    num_blocks_rows = arr.shape[0] // block_size
    num_blocks_cols = arr.shape[1] // block_size
    if num_blocks_rows == 0 or num_blocks_cols == 0:
        raise ValueError("Block size is too large for the input data dimensions.")
    blocks = arr[:num_blocks_rows * block_size,
    :num_blocks_cols * block_size].reshape(num_blocks_rows, block_size,
                                           num_blocks_cols, block_size)
    
    with np.testing.suppress_warnings() as sup:
        sup.filter(RuntimeWarning, "Mean of empty slice")
        _mean = np.nanmean(blocks, axis=(1, 3))
    with np.testing.suppress_warnings() as sup:
        sup.filter(RuntimeWarning, "Degrees of freedom <= 0 for slice")
        _std = np.nanstd(blocks, axis=(1, 3))
    enl = np.divide(_mean ** 2, _std ** 2,
                    out=np.full_like(_mean, fill_value=np.nan), where=_std != 0)
    out_arr = np.zeros((num_blocks_rows, num_blocks_cols))
    out_arr[:num_blocks_rows, :num_blocks_cols] = enl
    if not return_arr:
        if len(enl[~np.isnan(enl)]) == 0:
            return None
        out_median = np.nanmedian(out_arr).item()
        return round(out_median, decimals)
    else:
        return out_arr


def calc_performance_estimates(
        files: list[str],
        decimals: int = 2
):
    """
    Calculates the performance estimates specified in CARD4L NRB 1.6.9 for all noise power images if available.
    
    Parameters
    ----------
    files:
        List of paths pointing to the noise power images the estimates should be calculated for.
    decimals:
        Number of decimal places to round the calculated values to. Default is 2.
    
    Returns
    -------
    out: dict
        Dictionary containing the calculated estimates for each available polarization.
    """
    out = {}
    for f in files:
        pol = re.search('np-([vh]{2})', f).group(1).upper()
        with Raster(f) as ras:
            arr = ras.array()
            # The following need to be of type float, not numpy.float32 in order to be JSON serializable
            _min = float(np.nanmin(arr))
            _max = float(np.nanmax(arr))
            _mean = float(np.nanmean(arr))
            del arr
        out[pol] = {'minimum': round(_min, decimals),
                    'maximum': round(_max, decimals),
                    'mean': round(_mean, decimals)}
    return out


def evaluate_types(
        meta: Any
) -> None:
    """
    Evaluate the types of a metadata dictionary as extracted by e.g.
    :func:`s1ard.metadata.extract.meta_dict`.
    Currently allowed:
    
    - :py:obj:`bool`
    - :py:obj:`int`
    - :py:obj:`float`
    - :py:obj:`str`
    - :py:obj:`None`
    - :py:obj:`datetime.datetime`
    
    Parameters
    ----------
    meta
        the metadata dictionary to be evaluated.

    Raises
    ------
    TypeError

    """
    collector = dict()
    
    def get_typing(item=meta, collector=None, key=''):
        fill = '' if key == '' else '.'
        if isinstance(item, dict):
            return {k: get_typing(v, key=f'{key}{fill}{k}', collector=collector)
                    for k, v in item.items()}
        elif isinstance(item, list):
            return [get_typing(v, key=f'{key}{fill}{k}', collector=collector)
                    for k, v in enumerate(item)]
        else:
            t = type(item)
            # isinstance(np.float64(3.), float) == True
            allowed = [bool, int, float, str, type(None), None, datetime]
            if t not in allowed and collector is not None:
                collector[key] = t
            return t
    
    get_typing(item=meta, collector=collector)
    if len(collector) > 0:
        display = '\n'.join([f'{k}: {v}' for k, v in collector.items()])
        raise TypeError(f'the passed metadata dictionary contains unsupported types:\n'
                        f'{display}')


def get_header_size(tif: str) -> int:
    """
    Gets the header size of a GeoTIFF file in bytes.
    The code used in this function and its helper function `_get_block_offset` were extracted from the following
    source:
    
    https://github.com/OSGeo/gdal/blob/master/swig/python/gdal-utils/osgeo_utils/samples/validate_cloud_optimized_geotiff.py
    
    Copyright (c) 2017, Even Rouault
    
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:
    
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
    
    Parameters
    ----------
    tif:
        A path to a GeoTIFF file of the currently processed ARD product.

    Returns
    -------
    header_size:
        The size of all IFD headers of the GeoTIFF file in bytes.
    """
    
    def _get_block_offset(band):
        blockxsize, blockysize = band.GetBlockSize()
        for y in range(int((band.YSize + blockysize - 1) / blockysize)):
            for x in range(int((band.XSize + blockxsize - 1) / blockxsize)):
                block_offset = band.GetMetadataItem('BLOCK_OFFSET_%d_%d' % (x, y), 'TIFF')
                if block_offset:
                    return int(block_offset)
        return 0
    
    details = {}
    ds = gdal.Open(tif)
    main_band = ds.GetRasterBand(1)
    ovr_count = main_band.GetOverviewCount()
    
    block_offset = _get_block_offset(band=main_band)
    details['data_offsets'] = {}
    details['data_offsets']['main'] = block_offset
    for i in range(ovr_count):
        ovr_band = ds.GetRasterBand(1).GetOverview(i)
        block_offset = _get_block_offset(band=ovr_band)
        details['data_offsets']['overview_%d' % i] = block_offset
    
    headers_size = min(details['data_offsets'][k] for k in details['data_offsets'])
    if headers_size == 0:
        headers_size = gdal.VSIStatL(tif).size
    return headers_size



import os
import sys
from osgeo import gdal
import spatialist
import pyroSAR
import cesard


def keyval_check(
        key: str,
        val: str,
        allowed_keys: list[str]
) -> str | None:
    """
    Check and clean up key,value pairs while parsing a config file.
    
    Parameters
    ----------
    key:
        the parameter key
    val:
        the parameter value
    allowed_keys:
        a list of allowed keys
    """
    if key not in allowed_keys:
        msg = f"Parameter '{key}' is not allowed; should be one of {allowed_keys}"
        raise ValueError(msg)
    
    val = val.replace('"', '').replace("'", "")
    if val in ['None', 'none', '']:
        val = None
    return val


def validate_options(
        k: str,
        v: str,
        options: dict[str, list[str]]
) -> None:
    """
    Validate a configuration option against a set of allowed options.
    
    Parameters
    ----------
    k:
        the configuration key
    v:
        the configuration value
    options:
        the configuration options

    Returns
    -------

    """
    if k not in options:
        return
    if isinstance(v, list):
        for item in v:
            validate_options(k, item, options)
    else:
        msg = "Parameter '{}': expected value(s) to be one of {}; got '{}' instead"
        assert v in options[k], msg.format(k, options[k], v)


def validate_value(
        k: str,
        v: str | None | list[str]
) -> None:
    """
    Validate the value of a configuration option.
    
    Parameters
    ----------
    k:
        the configuration key
    v:
        the configuration value

    Returns
    -------

    """
    
    def val_aoi_geometry(x):
        return x is None or os.path.isfile(x)
    
    def val_aoi_tiles(x):
        return x is None or (isinstance(x, str) and len(x) == 5)
    
    def val_work_dir(x):
        return x is not None and os.path.isdir(v) and os.access(v, os.W_OK)
    
    validators = {'aoi_geometry': (val_aoi_geometry,
                                   'must be None or an existing file'),
                  'aoi_tiles': (val_aoi_tiles,
                                'must be None or a string of length 5'),
                  'work_dir': (val_work_dir,
                               'must be an existing, writable directory')}
    if k not in validators.keys():
        return
    if isinstance(v, list):
        for item in v:
            validate_value(k, item)
    else:
        validator, condition = validators[k]
        if not validator(v):
            msg = "Parameter '{}': value '{}' did not pass validation ({})."
            raise ValueError(msg.format(k, v, condition))


def version_dict() -> dict[str, str]:
    """
    Get the versions of used packages

    Returns
    -------
        a dictionary containing the versions of relevant python packages. Keys:
        
        - python
        - gdal
        - spatialist
        - pyrosar
        - cesard
    """
    out = {
        'python': sys.version,
        'gdal': gdal.__version__,
        'spatialist': spatialist.__version__,
        'pyrosar': pyroSAR.__version__,
        'cesard': cesard.__version__
    }
    return out



import pytest
import requests
from cesard.metadata.mapping import URL, DEM_MAP


def url_recursive(key, mapping, parent_key=None):
    key_info = f"{parent_key}.{key}" if parent_key else key
    if isinstance(mapping[key], dict):
        for k, v in mapping[key].items():
            url_recursive(k, mapping[key], key_info)
    else:
        value = mapping[key]
        if value is not None and value.startswith('https'):
            print(key_info)
            response = requests.get(
                url=value,
                headers={"User-Agent": "Mozilla/5.0"},
                allow_redirects=True
            )
            assert response.status_code == 200


@pytest.mark.parametrize('key', URL.keys())
def test_url(key):
    url_recursive(key=key, mapping=URL)


@pytest.mark.parametrize('key', DEM_MAP.keys())
def test_url_dem(key):
    url_recursive(key=key, mapping=DEM_MAP)



import re
import itertools
from lxml import html
from spatialist.vector import Vector, wkt2vector, bbox
from spatialist.auxil import utm_autodetect
from pyroSAR.drivers import ID
from cesard.ancillary import get_max_ext, buffer_min_overlap, get_kml, combine_polygons
from osgeo import ogr


def tile_from_aoi(
        vector: Vector | list[Vector],
        epsg: int | list[int] | None = None,
        strict: bool = True,
        return_geometries: bool = False,
        tilenames: list[str] | None = None
) -> list[str | Vector]:
    """
    Return a list of MGRS tile IDs or vector objects overlapping one or multiple areas of interest.
    
    Parameters
    -------
    vector:
        The vector object(s) to read. CRS must be EPSG:4236.
    epsg:
        Define which EPSG code(s) are allowed for the tile selection.
        If None, all tile IDs are returned regardless of projection.
    strict:
        Strictly only return the names/geometries of the overlapping tiles in the target projection
        or also allow reprojection of neighbouring tiles?
        In the latter case a tile name takes the form <tile ID>_<EPSG code>, e.g. `33TUN_32632`.
        Only applies if argument `epsg` is of type `int` or a list with one element.
    return_geometries:
        return a list of :class:`spatialist.vector.Vector` geometry objects (or just the tile names)?
    tilenames:
        an optional list of MGRS tile names to limit the selection
    
    Returns
    -------
        A list of unique MGRS tile IDs or :class:`spatialist.vector.Vector`
        objects with an attribute `mgrs` containing the tile ID.
    """
    if isinstance(epsg, int):
        epsg = [epsg]
    sortkey = None
    if return_geometries:
        sortkey = lambda x: x.mgrs
    kml = get_kml()
    with Vector(kml, driver='KML') as vec_kml:
        tiles = []
        with combine_polygons(vector, multipolygon=True) as vec_aoi:
            feature = vec_aoi.getFeatureByIndex(0)
            geom = feature.GetGeometryRef()
            vec_kml.layer.SetSpatialFilter(geom)
            feature = geom = None
            if tilenames is not None:
                values = ", ".join([f"'{x}'" for x in tilenames])
                sql_where = f"Name IN ({values})"
                vec_kml.layer.SetAttributeFilter(sql_where)
            for tile in vec_kml.layer:
                tilename = tile.GetField('Name')
                attrib = description2dict(tile.GetField('Description'))
                epsg_target = None
                if epsg is not None and attrib['EPSG'] not in epsg:
                    if len(epsg) == 1 and not strict:
                        epsg_target = int(epsg[0])
                        tilename += '_{}'.format(epsg_target)
                    else:
                        continue
                if return_geometries:
                    wkt = multipolygon2polygon(attrib['UTM_WKT'])
                    geom = wkt2vector_regrid(wkt=wkt,
                                             epsg_in=attrib['EPSG'],
                                             epsg_out=epsg_target)
                    geom.mgrs = tilename
                    field_defn = ogr.FieldDefn("mgrs", ogr.OFTString)
                    geom.layer.CreateField(field_defn)
                    feature = geom.getFeatureByIndex(0)
                    feature.SetField("mgrs", tilename)
                    geom.layer.SetFeature(feature)
                    tiles.append(geom)
                else:
                    tiles.append(tilename)
            tile = None
            geom = None
            layer = None
    return sorted(tiles, key=sortkey)


def aoi_from_tile(
        tile: str | list[str]
) -> Vector | list[Vector]:
    """
    Extract one or multiple MGRS tiles from the global Sentinel-2 tiling
    grid and return it as a :class:`~spatialist.vector.Vector` object.
    
    Parameters
    ----------
    tile:
        The MGRS tile ID(s) that should be extracted and returned as a vector object.
        Can also be expressed as <tile ID>_<EPSG code> (e.g. `33TUN_32632`). In this case the geometry
        of the tile is reprojected to the target EPSG code, its corner coordinates rounded to multiples
        of 10, and a new :class:`~spatialist.vector.Vector` object created.
    
    Returns
    -------
        either a single object or a list depending on `tile`
    """
    kml = get_kml()
    if isinstance(tile, str):
        tile = [tile]
    
    tilenames = []
    epsg_codes = []
    pattern = '([A-Z0-9]{5})_?([0-9]+)?'
    for i in tile:
        tilename, epsg = re.search(pattern, i).groups()
        tilenames.append(tilename)
        epsg_codes.append(epsg)
    
    values = ", ".join([f"'{x}'" for x in tilenames])
    sql_where = f"Name IN ({values})"
    out = []
    with Vector(kml, driver='KML') as vec:
        vec.layer.SetAttributeFilter(sql_where)
        for i, feat in enumerate(vec.layer):
            tilename = feat.GetField('Name')
            attrib = description2dict(feat.GetField('Description'))
            wkt = multipolygon2polygon(attrib['UTM_WKT'])
            epsg_target = epsg_codes[i]
            geom = wkt2vector_regrid(wkt=wkt,
                                     epsg_in=attrib['EPSG'],
                                     epsg_out=epsg_target)
            geom.mgrs = tilename
            field_defn = ogr.FieldDefn("mgrs", ogr.OFTString)
            geom.layer.CreateField(field_defn)
            feature = geom.getFeatureByIndex(0)
            feature.SetField("mgrs", tilename)
            geom.layer.SetFeature(feature)
            out.append(geom)
    if len(out) == 1:
        return out[0]
    else:
        return out


def description2dict(
        description: str
) -> dict[str, str | int]:
    """
    Convert the HTML description field of the MGRS tile KML file to a dictionary.
    
    Parameters
    ----------
    description:
        The plain text of the `Description` field
    
    Returns
    -------
        A dictionary with keys 'TILE_ID', 'EPSG', 'MGRS_REF', 'UTM_WKT' and 'LL_WKT'.
        The value of field 'EPSG' is of type integer, all others are strings.
    """
    attrib = html.fromstring(description)
    attrib = [x for x in attrib.xpath('//tr/td//text()') if x != ' ']
    attrib = dict(zip(attrib[0::2], attrib[1::2]))
    attrib['EPSG'] = int(attrib['EPSG'])
    return attrib


def aoi_from_scene(
        scene: ID,
        multi: bool = True,
        percent: int | float = 1
) -> list[dict[str, dict[str, float] | int]]:
    """
    Get processing AOIs for a SAR scene. The MGRS grid requires a SAR
    scene to be geocoded to multiple UTM zones depending on the overlapping
    MGRS tiles and their projection. This function returns the following
    for each UTM zone group:
    
    - the extent in WGS84 coordinates (key `extent`)
    - the extent in UTM coordinates (key `extent_utm`)
    - the EPSG code of the UTM zone (key `epsg`)
    
    A minimum overlap of the AOIs with the SAR scene is ensured by buffering
    the AOIs if necessary. The minimum overlap can be controlled with
    parameter `percent`.
    
    Parameters
    ----------
    scene:
        the SAR scene object
    multi:
        split into multiple AOIs per overlapping UTM zone or just one AOI
        covering the whole scene. In the latter case the best matching UTM
        zone is auto-detected
        (using function :func:`spatialist.auxil.utm_autodetect`).
    percent:
        the minimum overlap in percent of each AOI with the SAR scene.
        See function :func:`cesard.ancillary.buffer_min_overlap`.

    Returns
    -------
        a list of dictionaries with keys `extent`, `extent_utm`, `epsg`
    """
    out = []
    if multi:
        # extract all overlapping tiles
        with scene.geometry() as geom:
            tiles = tile_from_aoi(vector=geom, return_geometries=True)
        
        # group tiles by UTM zone
        def fn(x):
            return x.getProjection(type='epsg')
        
        for zone, group in itertools.groupby(tiles, lambda x: fn(x)):
            geometries = list(group)
            # get UTM EPSG code
            epsg = geometries[0].getProjection(type='epsg')
            # get maximum extent of tile group
            ext_utm = get_max_ext(geometries=geometries)
            del geometries
            with bbox(ext_utm, epsg) as geom1:
                # ensure a minimum overlap between AOI and pre-processed scene
                with scene.geometry() as geom2:
                    geom2.reproject(epsg)
                    # 60 m to keep aligned to MGRS tile size and overlaps
                    # see ancillary.check_spacing
                    with buffer_min_overlap(geom1=geom1, geom2=geom2,
                                            percent=percent, step=60) as buffered:
                        ext_utm = buffered.extent
                # convert extent to EPSG:4326
                geom1.reproject(projection=4326)
                ext = geom1.extent
            out.append({'extent': ext, 'epsg': epsg,
                        'extent_utm': ext_utm})
    else:
        with scene.bbox() as geom:
            ext = geom.extent
            # auto-detect UTM zone
            epsg = utm_autodetect(geom, 'epsg')
            # get all tiles, reprojected to the target UTM zone if necessary
            tiles = tile_from_aoi(vector=geom, epsg=epsg,
                                  return_geometries=True, strict=False)
        ext_utm = get_max_ext(geometries=tiles)
        del tiles
        out.append({'extent': ext, 'epsg': epsg,
                    'extent_utm': ext_utm})
    return out


def multipolygon2polygon(wkt: str) -> str:
    """
    Convert a MultiPolygon WKT with one Polygon to a simple Polygon WKT.
    The Sentinel-2 KML grid file stores all geometries as MultiPolygons.
    This function simply converts the geometries to simple Polygons.
    Not all geometries in the KML file have been checked. In case there are
    ever multiple Polygons in one MultiPolygon, an RuntimeError is raised.
    All other geometries are returned as is.
    
    Parameters
    ----------
    wkt:
        A geometry WKT representation.

    Returns
    -------
        the output WKT geometry. Either the original geometry or
        a Polygon extracted from a single-Polygon MultiPolygon.
    """
    geom1 = ogr.CreateGeometryFromWkt(wkt)
    poly_type = geom1.GetGeometryType()
    poly_num = geom1.GetGeometryCount()
    if poly_type == ogr.wkbMultiPolygon:
        if poly_num == 1:
            geom2 = geom1.GetGeometryRef(0).Clone()
            wkt = geom2.ExportToWkt()
            geom2 = None
        else:
            raise RuntimeError('got a MultiPolygon with multiple Polygons')
    geom1 = None
    return wkt


def wkt2vector_regrid(
        wkt: str,
        epsg_in: int,
        epsg_out: int | None = None
) -> Vector:
    """
    Convert a WKT geometry to a :class:`spatialist.vector.Vector` object and
    optionally reproject and regrid it.

    Parameters
    ----------
    wkt:
        the WKT string
    epsg_in:
        the EPSG code for the CRS of `wkt`
    epsg_out:
        and optional target CRS to reproject the geometry

    Returns
    -------
        the geometry object
    
    See Also
    --------
    spatialist.vector.wkt2vector
    """
    if epsg_out is None:
        return wkt2vector(wkt, epsg_in)
    else:
        with wkt2vector(wkt, epsg_in) as tmp:
            tmp.reproject(epsg_out)
            ext = tmp.extent
            for k, v in ext.items():
                ext[k] = round(v / 10) * 10
        return bbox(ext, crs=epsg_out)
